from __future__ import print_function
import dash
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output, State
from functools import wraps
import json
import numpy as np
import pandas as pd
import os
import plotly.figure_factory as ff
import plotly.graph_objs as go
import plotly.tools as tools
from scipy import stats

from pygcam.log import getLogger
from pygcam.mcs.analysis import getCorrDF
from pygcam.config import getConfig, DEFAULT_SECTION, getParam, setParam, setSection, getSections
from pygcam.mcs.Database import getDatabase
from pygcam.gui.widgets import dataStore
from pygcam.gui.styles import getColor, getStyle, updateStyle, getFont

_logger = getLogger(__name__)

CORR_STEP = 100

Oct16 = False       # special mode for specific presentation

def projectsWithDatabases():
    withDatabases = []

    for project in getSections():
        dbPath = getParam('MCS.DbPath', section=project, raiseError=False)
        if dbPath and os.path.exists(dbPath):
            withDatabases.append(project)

    return withDatabases

# Render slider labels in a consistent font and size
def sliderLabel(value):
    return {'label': str(value),
            'style': {'font-size': 10, 'font-family': 'Lato'}}

def cached(func):
    """
    Simple decorator to cache results keyed on method args plus project name.
    Note that this is not general purpose, but specialized for the McsData class.
    """
    cache = {}

    #@wraps  # keeps the name and doc string of wrapped function intact
    def wrapper(*args, **kwargs):
        self = args[0]
        key = (func.__name__, self.project, args[1:])

        try:
            return cache[key]
        except KeyError:
            cache[key] = result = func(*args, **kwargs)
            return result

    return wrapper

class McsData(object):
    def __init__(self, app):
        getConfig()

        # include only projects that have MCS databases
        self.projects = projectsWithDatabases()

        self.app = app
        self.db = None
        self.project = None
        self.simId = None
        self.inputs = None
        self.inputsDF = None
        self.resultDict = None
        self.histogramContext = None
        self.paraCoordsVars = None
        self.resetSliderButtonClicks = None
        self.selectedResults = None
        self.corrDF = None

    def readMetaData(self, project=None):
        if project:
            setParam('GCAM.DefaultProject', project, DEFAULT_SECTION)
            setSection(project)
        else:
            project = getParam('GCAM.DefaultProject')

        self.project = project
        self.inputsDF = None
        self.resultDict = None
        self.selectedResults = None
        self.corrDF = None

        self.simId = None

        if self.db:
            self.db.close()

        db = self.db = getDatabase()
        self.inputs = db.getInputs()

    @cached
    def getParameterValues(self, simId):
        db = self.db
        _logger.debug('Reading inputDF...',)
        inputsDF = db.getParameterValues2(simId)

        # Drop any inputs with names ending in '-linked' since they're redundant
        linked = filter(lambda s: s.endswith('-linked'), inputsDF.columns)
        if linked:
            inputsDF = inputsDF.drop(linked, axis=1, inplace=False)

        # Handle special case of ramp-index by rounding to integer
        col = 'ramp-index'
        if col in inputsDF.columns:
            inputsDF[col] = inputsDF[col].apply(round)

        numParams = inputsDF.shape[0] * inputsDF.shape[1]
        _logger.info('%d parameter values read' % numParams)
        return inputsDF

    @cached
    def getOutputsWithValues(self, simId, scenario):
        outputs = self.db.getOutputsWithValues(simId, scenario)
        return outputs

    @cached
    def getOutValues(self, simId, scenario, resultName, limit=None):
        """
        Just a layer over db.getOutValues to provide caching of results, and
        to convert result from a DataFrame to a Series
        """
        df = self.db.getOutValues(simId, scenario, resultName, limit=limit)
        series = df[resultName]

        # TBD: generalize this
        # Convert these so we can display 0.5 as 50%
        if resultName == 'percent-change':
            series *= 100

        return series

    # Deprecated?
    # def getResults(self, simId, scenarioList, resultList=None, limit=None):
    #     '''
    #     Get the results for the given result names, or, if none are
    #     specified, for the results identified when at instantiation.
    #     Results are cached and thus read only once from the database.
    #
    #     :param scenarioList: (iterable of str) the scenarios for which
    #        to get results
    #     :param resultList: (iterable of str) the (scalar) results to read
    #     :return: (DataFrame) rows contain result name, scenario name, and value
    #     '''
    #
    #     db = self.db
    #     resultDict = {}
    #     key = None
    #
    #     if resultList is None and limit is None:
    #         key = ('getResults', simId, frozenset(scenarioList))
    #         cached = getFromCache(key)
    #         if cached is not None:
    #             return cached
    #
    #     resultList = resultList or db.getOutputs()
    #     scenarioList = scenarioList or db.scenariosWithResults(simId)
    #
    #     for scenario in scenarioList:
    #         resultDF = resultDict.get(scenario)
    #
    #         for resultName in resultList:
    #             # returns DF with 'trialNum' as index, 'value' holds float value
    #             if resultDF is None or resultName not in resultDF.columns:
    #                 values = self.getOutValues(simId, scenario, resultName, limit=limit)
    #                 if values is not None:
    #                     resultDF[resultName] = values
    #
    #         resultDict[scenario] = resultDF
    #
    #     self.resultDict = resultDict
    #     return storeInCache(key, resultDict)

    def projectChooser(self):
        layout = dcc.Dropdown(id='project-chooser',
                              options=[{'label':name, 'value':name} for name in self.projects],
                              value=getParam('GCAM.DefaultProject'))
        return layout

    def simChooser(self, optionsOnly=False):
        sims = self.db.getSims()
        if sims is None:
            sims = []

        def simLabel(sim):
            desc  = sim.description or '[no description]'
            label = "%d. %s (%d trials)" % (sim.simId, desc, sim.trials)
            return label

        options = [{'label':simLabel(sim), 'value':sim.simId} for sim in sims]

        if optionsOnly:
            return options

        layout = dcc.Dropdown(
            id='sim-chooser',
            options=options,
            value=sims[0].simId if sims else None)
        return layout

    def scenarioChooser(self, simId=None, multi=False, optionsOnly=False):
        scenarios = [] if simId is None else (self.db.scenariosWithResults(simId) or [])
        options = [{'label': s, 'value': s} for s in scenarios]

        if optionsOnly:
            _logger.debug("scenarioChooser returning %s" % options)
            return options

        layout = dcc.Dropdown(
            id='scenario-chooser',
            options=options,
            value=options[0]['value'] if options else '',
            multi=multi
        )
        _logger.debug("scenarioChooser returning %s" % layout)
        return layout

    def inputChooser(self, multi=False, optionsOnly=False):
        options = [{'label': name, 'value': name} for name in self.inputs]

        if optionsOnly:
            return options

        layout = dcc.Dropdown(
            id='input-chooser',
            options=options,
            multi=multi,
            value=[])
        return layout

    def outputChooser(self, simId=None, scenario=None, optionsOnly=False):
        _logger.debug("outputChooser: simId={}, scenario={}".format(simId, scenario))
        outputs = self.getOutputsWithValues(simId, scenario) \
            if scenario and (simId is not None) else []

        options = [{'label': name, 'value': name} for name in outputs]

        if optionsOnly:
            return options

        layout = dcc.Dropdown(
            id='output-chooser',
            options=options,
            multi=False,
            value=options[0]['value'] if options else '')
        return layout

    def multiOutputChooser(self, simId=None, scenario=None, optionsOnly=False):
        outputs = self.getOutputsWithValues(simId, scenario) \
            if scenario and (simId is not None) else []

        options = [{'label': name, 'value': name} for name in outputs]

        if optionsOnly:
            return options

        layout = dcc.Checklist(
            id='multi-output-chooser',
            options=options,
            values=outputs, # pre-select everything
            labelStyle={'display': 'inline-block', 'margin': '6px'})

        return layout

    def distPlot(self, simId, scenario, outputName, sliderInfo, distOptions, selectedData):
        """
        Generate the plot data for the distribution plot. Return three items:
        the plot data, title, and annotations (might be empty list).
        """
        annotations = []
        colors = []

        values = self.getOutValues(simId, scenario, outputName)

        if values is None:
            title = 'Please select a model output to plot'
            plotData = [dict(type='bar', x=[], y=[],
                             name=outputName, showlegend=False,)]
            return plotData, title, annotations

        title = '%s for scenario %s' % (outputName, scenario)

        bins = min(100, len(values)) or 1 # 1 for corner case of no data; bins must be > 0

        # Generate histogram data to be able to color bars directly
        counts, edges = np.histogram(values, bins, density=True)
        barCount = len(counts)

        # Plot the counts on x-axis at the mean of each pair of edges
        barValues = [np.mean(edges[i:i + 1]) for i in range(barCount)]

        if sliderInfo:
            minQ, maxQ = sliderInfo
            minX = values.quantile(q=minQ / 100.0, interpolation='linear')
            maxX = values.quantile(q=maxQ / 100.0, interpolation='linear')
        else:
            minX, maxX = selectedData['range']['x'] if selectedData else (edges[0], edges[-1])

        # Sub-select trials based on selection range to show in parallel coords plot
        self.selectedResults = values[values.between(minX, maxX, inclusive=True)]

        # highlight selected bars by desaturating non-selected items
        active   = getColor('DistBarsActive')
        inactive = getColor('DistBarsInactive')
        for i in range(barCount):
            colors.append(active if minX <= barValues[i] <= maxX else inactive)

        distPlot = ff.create_distplot([values], [outputName], bin_size=1,
                                      show_hist=False, show_rug=False, curve_type='kde')

        # TBD: generalize this
        if outputName == 'percent-change':
            tickvalues = ['%d%%' % int(value) for value in barValues]
        else:
            tickvalues = ['%.2f' % value for value in barValues]

        plotData = [dict(type='bar',
                         x=barValues,
                         y=list(counts),
                         name=outputName,
                         histnorm='probability density',
                         marker=dict(color=colors),
                         showlegend=False,
                         text=tickvalues,
                         hoverinfo='text'
                         )]

        if 'kde' in distOptions:
            distData = distPlot.data[0]
            plotData.append(dict(type='scatter',
                                 x=distData['x'],
                                 y=distData['y'],
                                 marker=dict(color=getColor('KDE')),
                                 hoverinfo='skip', # prevents hover events
                                 showlegend=False,
                                 ))

        showMean = 'mean' in distOptions
        showMedian = 'median' in distOptions

        if showMean or showMedian:
            info = values.describe()
            maxY = max(counts)
            labelSize = 12
            bgcolor = getColor('PlotBg')

            mean   = info['mean']
            median = info['50%']

            def lineData(name, x, maxY, color):
                d = dict(type='scatter', mode='lines',
                         name=name,
                         x=[x, x], y=[0.0000001, maxY],
                         marker=dict(color=color))
                return d

            def notation(x, y, xanchor, color):
                # TBD: generalize this
                if outputName == 'percent-change':
                    text = '%d%%' % x
                else:
                    text = '%.2f' % x

                d = dict(x=x, y=y,
                         text=text,
                         xanchor=xanchor, yanchor='top',
                         showarrow=False,
                         bgcolor=bgcolor, opacity=0.7,
                         font=dict(family='Lato', size=labelSize, color=color))
                return d

            if showMean:
                color = getColor('MeanLineColor')
                xanchor = 'right' if mean <= median else 'left'
                plotData.append(lineData('mean', mean, maxY, color))
                annotations.append(notation(mean, 0.95 * maxY, xanchor, color))

            if showMedian:
                color = getColor('MedianLineColor')
                xanchor = 'right' if median < mean else 'left'
                plotData.append(lineData('median', median, maxY, color))
                annotations.append(notation(median, maxY, xanchor, color))

        return plotData, title, annotations

    @cached
    def getCorrDF(self, simId, scenario, resultName):
        results = self.getOutValues(simId, scenario, resultName)
        inputsDF = self.getParameterValues(simId)
        inputsDF = inputsDF.iloc[results.index]      # select only trials for which we have results

        corrDF = getCorrDF(inputsDF, results)
        return corrDF

    @cached
    def getCorrByTrials(self, simId, scenario, resultName):
        results  = self.getOutValues(simId, scenario, resultName)
        inputsDF = self.getParameterValues(simId)

        # # TBD: a hack to examine failures
        # idx = set(inputsDF.index[:max(results.index)]) - set(results.index)
        # failures = inputsDF.iloc[list(idx)]
        #   failures.to_csv('/Users/rjp/tmp/failures.csv')

        inputsDF = inputsDF.iloc[results.index]      # select only trials for which we have results

        paramsToShow = 10
        fullDF = getCorrDF(inputsDF, results)
        topParams = list(fullDF[:paramsToShow].index)

        colsToDrop = set(inputsDF.columns) - set(topParams)
        inputsDF = inputsDF.drop(colsToDrop, axis=1, inplace=False) # not inplace since it's a slice of the original

        trialSteps = range(CORR_STEP, len(results), CORR_STEP)      # produce corrDF for increments of 100 trials
        trialSteps.append(len(results))                             # final value is for however many trials there were

        corrByTrials = pd.DataFrame()

        for count in trialSteps:
            corrDF = getCorrDF(inputsDF[:count], results[:count])
            corrDF['count'] = count
            corrByTrials = corrByTrials.append(corrDF)

        corrByTrials.reset_index(inplace=True)
        return corrByTrials


    def parcoordsPlot(self, simId, scenario, resultName, numVars):
        if simId is None or not (scenario and resultName):
            return ''

        result   = self.getOutValues(simId, scenario, resultName)
        inputsDF = self.getParameterValues(simId)
        inputsDF = inputsDF.iloc[result.index]      # select only trials for which we have results

        corrDF = self.getCorrDF(simId, scenario, resultName)
        varCount = min(numVars, inputsDF.shape[1])
        varNames = corrDF.index[:varCount]
        self.paraCoordsVars = list(varNames)

        def appendDim(series, name, constraint=None):
            d = dict(range=[round(min(series), 2),
                            round(max(series), 2)],
                     label=name,
                     values=series)

            if constraint:
                # constrain by selection from distribution plot
                d['constraintrange'] = constraint

            dimensions.append(d)

        dimensions = []
        for name in varNames:
            appendDim(inputsDF[name], name)

        selected = self.selectedResults
        appendDim(result, resultName, constraint=[min(selected), max(selected)])

        plotData = [go.Parcoords(
            line=dict(color=result,
                      colorscale='Jet',
                      showscale=True,
                      reversescale=False,
                      cmin=min(result),
                      cmax=max(result)),
            dimensions=dimensions,
            rangefont={'color': 'rgba(255,0,0,0)'}, # zero alpha => invisible
            )]

        layout = dict(autosize=True,
                      height=500,
                      margin=dict(l=75, r=40, b=35, t=35),
                      hovermode="closest",
                      title='',
                      showlegend=False,
                      legend=dict(font=dict(size=10), orientation='h'),
                      )
        figure = dict(data=plotData, layout=layout)
        return figure

    def tornadoPlot(self, simId, scenario, resultName, tornadoType,
                    sliderValue, selectedData):
        corrDF = self.getCorrDF(simId, scenario, resultName)

        if tornadoType == 'normalized':
            squared = corrDF.spearman ** 2
            corrDF['normalized'] = squared / squared.sum()
            corrDF['sign'] = 1
            corrDF.ix[(corrDF.spearman < 0), 'sign'] = -1
            corrDF['value'] = corrDF.normalized * corrDF.sign     # normalized squares with signs restored
            plotColumn = 'value'
            title = 'Normalized rank correlation'
        else:
            plotColumn = 'spearman'
            title = 'Rank correlation'

        varCount = min(20, len(corrDF)) # if sliderValue is None else len(corrDF[corrDF['abs'] >= sliderValue]))
        varNames = list(reversed(corrDF.index[:varCount]))
        values   = list(reversed(corrDF[plotColumn][:varCount]))

        plotData = go.Bar(
            orientation='h',
            x=values,
            y=varNames,
            width=0.6,
            marker=dict(
                color=getColor('TornadoPlotBar'),
                line=dict(
                    color=getColor('TornadoPlotLine'),
                    width=0),
            ),

            textposition='none',
            name='Rank correlation',
            hoverinfo='skip', # prevents hover events
        )

        layout = updateStyle('Plot',
                             title=title,
                             margin=updateStyle('PlotMargin', l=200),
                             xaxis=dict(range=[-1, 1]),
                             dragmode='select')

        figure = dict(data=[plotData], layout=layout)
        return figure

    def corrConvergencePlot(self, simId, scenario, resultName):
        df = self.getCorrByTrials(simId, scenario, resultName)
        params = list(df['paramName'].unique())
        traces = []

        count = max(df['count'])

        # create a trace for each parameter
        for paramName in params:
            paramData = df.query('paramName == "%s"' % paramName)
            trace = go.Scatter(x=list(paramData['count']),
                               y=list(paramData['spearman']),
                               name=paramName, mode='lines')
            traces.append(trace)

        # Round up to nearest integral value of CORR_STEP
        endX = count + (CORR_STEP - count % CORR_STEP)

        # I cannot center a narrower plot for reasons I don't understand
        layout = updateStyle('Plot',
                             width=None,
                             title='Correlation Convergence for %s' % resultName,
                             xaxis=dict(range=[CORR_STEP, endX]))
                             # margin=updateStyle('PlotMargin', l=40)

        figure = dict(data=traces, layout=layout)
        return figure

    def scatterPlots(self, simId, scenario, inputs, outputs):
        """
        Plot a set of small scatterplots showing correlation between chosen
        model inputs and outputs.

        :param simId: (int) simulation id
        :param scenario: (str) scenario name
        :param inputs: (iterable of str) names of inputs to plot
        :param outputs: (iterable of str) names of outputs to plot

        :return: dash 'figure' data.
        """
        allInputs = self.getParameterValues(simId)
        inputsDF  = allInputs[inputs]

        outputsDF = pd.DataFrame(columns=outputs)
        for output in outputs:
            outputsDF[output] = self.getOutValues(simId, scenario, output)

        inputsDF = inputsDF.iloc[outputsDF.index]      # select only trials for which we have results

        numIns  = len(inputs)
        numOuts = len(outputs)
        fig = tools.make_subplots(rows=numOuts, cols=numIns,
                                  vertical_spacing=0.01,
                                  shared_xaxes=True,
                                  shared_yaxes=True)
        layout = fig['layout']
        for row, output in enumerate(outputs):
            yAxisNum = row + 1
            xAxisNum = 0

            for col, input in enumerate(inputs):
                xAxisNum += 1

                trace = go.Scatter(
                    x=inputsDF[input],
                    y=outputsDF[output],
                    hoverinfo='skip',       # don't show or fire events
                    mode='markers',
                    marker=dict(size=1,
                                opacity=0.9,
                                color=getColor('ScatterPoints'))
                )

                fig.append_trace(trace, row+1, col+1)

                # Compute the names of the corresponding axes
                xaxis = "xaxis{}".format(xAxisNum)
                yaxis = "yaxis{}".format(yAxisNum)

                layout[xaxis].update(title=input,  showgrid=False, zeroline=False)
                layout[yaxis].update(title=output, showgrid=False, zeroline=False)

        layout.update(margin=dict(l=50, r=30, t=30, b=30),
                      height=140 * numOuts + 20,
                      #width= 130 * numIns,
                      showlegend=False,
                      font=dict(size=9,
                                family=getFont('PlotText')))
        return fig

    def distributionSectionLayout(self):
        # slider, distribution, checklist
        layout = [
            html.P([
                html.Span('Filter using the percentile slider or select a range in the histogram. Selected range: '),
                html.Span(id='percentile-text')
            ], style=getStyle('HelpText')),

            html.Div([
                html.Div([
                    dcc.RangeSlider(
                        id='dist-slider',
                        min=0,
                        max=100,
                        step=0.1,
                        marks={percentile: sliderLabel(str(percentile) + '%') for percentile in range(0, 101, 10)},
                        # marks={percentile: str(percentile) + '%' for percentile in range(0, 101, 10)},
                        # updatemode='drag', # too slow redrawing parcoords
                        value=[0, 100]),
                ], style={'width': '350',
                          'display': 'inline-block'}),

                # gap between slider and button
                html.Div(' ',
                         style={'width': 10, 'display': 'inline-block'}),

                html.Div([
                    html.Button('Reset',
                                id='dist-slider-reset-button',
                                style=getStyle('Button')
                                )
                ], style={'width': '55',
                          'display': 'inline-block'})

            ], style={'margin-left': 10}),

            html.Div(
                dcc.Graph(id='distribution'),
                style={'margin-top': 20}),

            dcc.Checklist(
                id='dist-options',
                options=[{'label': 'KDE', 'value': 'kde'},
                         {'label': 'Mean', 'value': 'mean'},
                         {'label': 'Median', 'value': 'median'}],
                values=['kde'],
                labelStyle={'display': 'inline-block', 'margin': '6px'})
        ]

        return layout

    def tornadoSectionLayout(self):
        layout = [

            html.P([
                html.Span('Adjust the number of variables to show by setting the correlation cut-off'),
                html.Span(id='corr-cutoff-text')
            ], style=updateStyle('HelpText', display='none')),   # getStyle('HelpText')),

            html.Div([
                html.Div([
                    dcc.Slider(
                        id='tornado-slider',
                        disabled=True,
                        min=0, max=0.4, step=0.01,
                        marks={value: sliderLabel(value) for value in np.arange(0, 0.4, 0.1)},
                        updatemode='drag',
                        value=0.01),
                ], style={
                    'width': 300,
                    # 'display': 'inline-block'
                    'display': 'none'
                    }
                ),
            ]),

            html.Div(
                dcc.Graph(id='tornado'),
                style={'margin-top': 74}),  # 20}),

            dcc.RadioItems(
                id='tornado-type-chooser',
                options=[{'label': 'Rank correlation',
                          'value': 'spearman'},
                         {'label': 'Normalized',
                          'value': 'normalized'}],
                value='normalized',
                labelStyle={'display': 'inline-block', 'margin': '6px'})
        ]
        return layout

    def layout(self):
        layout = html.Div([
            dataStore(id='paraCoords-vars'),
            dataStore(id='dist-selected'),

            html.H1('Monte Carlo Simulation Explorer', className='title'),
            html.Table([
                html.Tr([
                    html.Th('Project'),
                    html.Th('Simulation'),
                    html.Th('Scenario'),
                    html.Th('Model Output'),
                ]),
                html.Tr([
                    html.Td(self.projectChooser()),
                    html.Td(self.simChooser()),
                    html.Td(self.scenarioChooser()),
                    html.Td(self.outputChooser())
                ])

            ], style={'width': '100%',
                      'border': '1px solid black',
                      'table-layout': 'fixed',
                      'background-color': getColor('ChooserBgColor')}),

            html.Div([
                # distribution cluster and tornado chart
                html.Div([
                    html.Div(self.distributionSectionLayout(),
                             className='cell twocol'),

                    html.Div(self.tornadoSectionLayout(),
                             className='cell twocol')
                ], className='row'),

                # parallel coordinates section
                html.Div([
                    html.Div([
                        html.P(['Brush vertically over ranges to filter; click outside the selected range to reset. ',
                                'Drag variable names below to reorder. Slider sets the number of vars to plot'],
                           style=getStyle('HelpText')),

                        html.Div([
                            dcc.Slider(
                                id='paraCoords-slider',
                                min=1, max=10, step=1,
                                marks={value: sliderLabel(value) for value in range(1, 11, 1)},
                                updatemode='drag',
                                value=6),
                        ], style={'margin-bottom': 30,
                                  'margin-left':  'auto',
                                  'margin-right': 'auto',
                                  'width': 400}),

                        html.Div(dcc.Graph(id='paraCoords')),

                    ], className='cell onecol')
                ], className='row'),

                # Correlation convergence section
                html.Div([
                    html.Div([
                        html.Span('Correlation convergence', style=getStyle('Label')),
                        html.Div('', style={'height': 15}),
                        html.Div(dcc.Graph(id='corr-convergence')),
                    ], className='cell onecol')
                ], className='row'),

                # Scatterplot section
                html.Div([
                    html.Div([
                        html.Div([
                            html.Div(style={'height': 15}),
                            html.Span('Model Inputs (random variables)', style=getStyle('Label')),
                            html.Div(style={'height': 15}),
                            self.inputChooser(multi=True)
                        ], style=updateStyle('Chooser', width=900)),

                        html.Div(style={'height':15}),

                        html.Div([
                            html.Span('Model Outputs', style=getStyle('Label')),
                            self.multiOutputChooser()
                        ], style=updateStyle('Chooser', width=900)),

                        html.Div([
                            html.Button('Create scatterplot',
                                        id='scatterplot-button',
                                        style=getStyle('Button')
                                        )
                        ], style={'margin': 5}),

                        html.Div([
                            dcc.Graph(id='scatter-matrix')
                        ], style=updateStyle('AutoMargins'))
                    ], className='cell onecol')
                ], className='row'),

            ], className='table',
                style={'width': 1200,
                       'margin-left': 'auto',
                       'margin-right': 'auto'}),

        ], style=getStyle('Page'))

        return layout

# getTimeSeries(self, simId, paramName, expList)


def generateDropdownDefaults(app, ids):
    """
    Generate dropdown default-setting using the 'callback' decorator function.

    :param app: a Dash app instance
    :param ids: (list(str)) ids of Dropdown controllers to generate callbacks for.
    :return: none
    """
    def dropdownDefault(options, current):
        if not options:
            return ''

        optionValues = [opt['value'] for opt in options]

        # if current value is legit, leave it, otherwise choose first item
        return current if current in optionValues else optionValues[0]

    for id in ids:
        _logger.debug("Generating default-setting callback for %s" % id)
        app.callback(Output(id, 'value'),
                     [Input(id, 'options')],
                     state=[State(id, 'value')])(dropdownDefault)

def _setup_werkzeug_log(level, host, port):
    import logging

    consoleFormat = getParam('GCAM.LogConsoleFormat')
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(consoleFormat))
    log = logging.getLogger('werkzeug')
    log.addHandler(handler)
    log.setLevel(level)
    log.propagate = False

    # Setting Flask.LogLevel > INFO quashes this useful message, so we restore it
    if log.level > logging.INFO:
        print('* Running on http://{}:{}/ (Press CTRL+C to quit)'.format(host, port))


def main(args):
    app = dash.Dash(name='mcs-explorer', csrf_protect=False)

    level = getParam('Flask.LogLevel')
    flaskLog = app.server.logger
    flaskLog.setLevel(level)
    _setup_werkzeug_log(level, args.host, args.port)

    # app.config.supress_callback_exceptions = True
    app.css.append_css({"external_url": "https://fonts.googleapis.com/css?family=Dosis:regular,bold|Lato"})
    app.css.append_css({"external_url": "https://codepen.io/plevin/pen/MvpeNV.css"})

    data = McsData(app)
    data.readMetaData()

    app.layout = data.layout()

    #
    # Callbacks
    #
    generateDropdownDefaults(app, ['scenario-chooser', 'output-chooser'])


    @app.callback(Output('output-chooser', 'options'),
                  [Input('project-chooser', 'value'),
                   Input('sim-chooser', 'value'),
                   Input('scenario-chooser', 'value')])
    def updateOutputChooser(project, simId, scenario):
        _logger.debug('updateOutputChooser(%s, %s, %s)' % (project, simId, scenario))
        layout = data.outputChooser(simId=simId, scenario=scenario, optionsOnly=True)
        return layout

    @app.callback(Output('multi-output-chooser', 'options'),
                  [Input('project-chooser', 'value'),
                   Input('sim-chooser', 'value'),
                   Input('scenario-chooser', 'value')])
    def updateMultiOutputChooser(project, simId, scenario):
        _logger.debug('updateMultiOutputChooser(%s, %s, %s)' % (project, simId, scenario))
        layout = data.multiOutputChooser(simId=simId, scenario=scenario, optionsOnly=True)
        return layout

    @app.callback(Output('scenario-chooser', 'options'),
                  [Input('project-chooser', 'value'),
                   Input('sim-chooser', 'value')])
    def updateScenarioChooser(project, simId):
        _logger.debug('updateScenarioChooser(%s, %s)' % (project, simId))
        layout = data.scenarioChooser(simId, multi=False, optionsOnly=True)

        return layout

    @app.callback(Output('sim-chooser', 'options'),
                  [Input('project-chooser', 'value')])
    def updateSimChooser(project):
        _logger.debug('updateSimChooser(%s)' % project)
        data.readMetaData(project)
        layout = data.simChooser(optionsOnly=True)
        return layout

    # TBD: finish writing these methods to store data in hidden div for
    # TBD: reset button and selection to simplify callback for drawing dist

    #
    # The next few callbacks use a hidden <div> to store
    # communicate the variables used in the parallel coords
    # figure with those in the output chooser.
    #

    @app.callback(Output('paraCoords-vars', 'children'),
                  [Input('paraCoords', 'figure')])
    def updateParaCoordsVars(figure):
        varNames = data.paraCoordsVars
        _logger.debug('updateParaCoordsVars vars=%s' % varNames)
        varNamesJSON = json.dumps(varNames)
        return varNamesJSON

    @app.callback(Output('input-chooser', 'value'),
                  [Input('paraCoords-vars', 'children')])
    def updateInputChooser(varNamesJSON):
        varNames = json.loads(varNamesJSON)
        _logger.debug('updateInputChooser(%s)' % varNames)
        return varNames

    @app.callback(Output('paraCoords', 'figure'),
                  [Input('project-chooser', 'value'),
                   Input('sim-chooser', 'value'),
                   Input('scenario-chooser', 'value'),
                   Input('output-chooser', 'value'),
                   Input('paraCoords-slider', 'value'),
                   Input('distribution', 'figure'),     # force this to plot last
                   ])
    def showParallelCoords(project, simId, scenario, resultName, varsToShow, figure):
        return data.parcoordsPlot(simId, scenario, resultName, varsToShow)

    @app.callback(Output('percentile-text', 'children'),
                  [Input('dist-slider', 'value')])
    def showDistSliderValues(sliderInfo):
        return '%d%%-%d%%' % tuple(sliderInfo)

    @app.callback(Output('tornado', 'figure'),
                  [Input('project-chooser', 'value'),
                   Input('sim-chooser', 'value'),
                   Input('scenario-chooser', 'value'),
                   Input('output-chooser', 'value'),
                   Input('tornado-slider', 'value'),
                   Input('tornado-type-chooser', 'value'),
                   Input('distribution', 'selectedData'),
                  ])
    def showTornado(project, simId, scenario, resultName,
                    sliderValue, tornadoType, selectedData):
        _logger.debug('showTornado(%s, %s, %s, %s, %s, %s, %s)' % (
            project, simId, scenario, resultName, sliderValue, tornadoType, selectedData))

        if not (project and scenario and resultName) or simId is None:
            return ''

        figure = data.tornadoPlot(simId, scenario, resultName, tornadoType,
                                  sliderValue, selectedData)
        return figure

    @app.callback(Output('distribution', 'figure'),
                  [Input('project-chooser', 'value'),
                   Input('sim-chooser', 'value'),
                   Input('scenario-chooser', 'value'),
                   Input('output-chooser', 'value'),
                   Input('dist-slider', 'value'),
                   Input('dist-options', 'values'),
                   Input('distribution', 'selectedData'),
                   ])
    def showDistribution(project, simId, scenario, outputName, sliderInfo, distOptions, selectedData):
        _logger.debug('showDistribution(%s, %s, %s, %s, %s)' % (
            project, simId, scenario, outputName, selectedData))

        # Clear selectedData if context changes
        context = (project, simId, scenario, outputName)
        if not data.histogramContext or context != data.histogramContext:
            _logger.debug('showDistribution: clearing context')
            selectedData = sliderInfo = None
            data.histogramContext = context

        # if called before values are known, return an empty plot
        if not (project and scenario and outputName) or simId is None:
            plotData = {'x': []}
            title = 'Please select a model output to plot'
            annotations = None
        else:
            plotData, title, annotations = data.distPlot(simId, scenario, outputName,
                                                         sliderInfo, distOptions,
                                                         selectedData)
        # TBD: generalize this
        if Oct16:
            if outputName == 'percent-change':
                xtitle = 'Change from baseline fuel'
                xticksuffix = '%'
            else:
                # latex formatting is broken, but HTML works fine
                # xtitle = '$g CO_2 MJ^{-1}$' if Oct16 else ''
                xtitle = 'g CO<sub>2</sub> MJ<sup>-1</sup>'
                xticksuffix = None
        else:
            xtitle = ''
            xticksuffix = ''

        layout = updateStyle('Plot',
                             title=title,
                             yaxis={'title': 'Probability density',
                                    'tickvals': []
                                    },
                             xaxis={'title': xtitle,
                                    'ticksuffix': xticksuffix,
                                    },
                             margin=getStyle('PlotMarginWithXTitle'),
                             dragmode='select',
                             showLegend=True,
                             legend=dict(x=0.0, y=1.0, bgcolor=getColor('PlotBg')))

        if annotations:
            layout['annotations'] = annotations

        figure = dict(data=plotData, layout=layout)
        return figure

    # dist -> dist slider
    @app.callback(Output('dist-slider', 'value'),
                  [Input('output-chooser', 'value'),
                   Input('distribution', 'selectedData'),
                   Input('dist-slider-reset-button', 'n_clicks')],
                  [State('sim-chooser', 'value'),
                   State('scenario-chooser', 'value')])
    def showPercentileSlider(resultName, selected, clicks, simId, scenario):
        _logger.debug('showPercentileSlider(%s, %s, %s)' % (resultName, selected, clicks))
        if clicks != data.resetSliderButtonClicks:            # recognize a new click
            data.resetSliderButtonClicks = clicks
            return (0, 100)

        if not selected or not resultName:
            return (0, 100)

        minX, maxX = selected['range']['x']
        values = data.getOutValues(simId, scenario, resultName)
        minQ = stats.percentileofscore(values, minX)
        maxQ = stats.percentileofscore(values, maxX)
        return [minQ, maxQ]

    # scatterplot matrix
    @app.callback(Output('scatter-matrix', 'figure'),
                  [Input('scatterplot-button', 'n_clicks')],
                  [State('sim-chooser', 'value'),
                   State('scenario-chooser', 'value'),
                   State('multi-output-chooser', 'values'),
                   State('input-chooser', 'value')])
    def showScatterMatrix(nclicks, simId, scenario, outputs, inputs):
        inputs = inputs or data.paraCoordsVars
        #outputs = outputs or data.getOutputsWithValues(simId, scenario)
        _logger.debug('showScatterMatrix(%s, %s, %s, %s)' % (simId, scenario, inputs, outputs))

        if not inputs or not outputs:
            return ''

        figure = data.scatterPlots(simId, scenario, inputs, outputs)
        return figure

    if True:
        # TBD: not needed in all cases; restore this as option later
        # correlation convergence plot
        @app.callback(Output('corr-convergence', 'figure'),
                      [Input('project-chooser', 'value'),
                       Input('sim-chooser', 'value'),
                       Input('scenario-chooser', 'value'),
                       Input('output-chooser', 'value')])
        def showCorrConvergence(project, simId, scenario, resultName):
            if not (project and scenario and resultName) or simId is None:
                return ''

            figure = data.corrConvergencePlot(simId, scenario, resultName)
            return figure

    app.run_server(debug=args.debug, threaded=False, host=args.host, port=args.port)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
from __future__ import print_function
import os

def main():
    cfg = os.path.expanduser('~/.pygcam.cfg')
    with open(cfg, 'r') as f:
        lines = f.readlines()

    match = filter(lambda line: line.startswith('GCAM.DefaultProject'), lines)
    if match:
        line = match[0]
        defaultProject = line.split('=')[1].strip()
        print(defaultProject)
    else:
        print('-- Not set --')

main()



'''
.. codeauthor:: Richard Plevin

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from __future__ import print_function
from ..subcommand import SubcommandABC

class GUICommand(SubcommandABC):

    def __init__(self, subparsers):
        kwargs = {'help' : '''Run the Graphical User Interface generated from the command-line interface.'''}

        super(GUICommand, self).__init__('gui', subparsers, kwargs, group='utils', guiSuppress=True)

    def addArgs(self, parser):
        parser.add_argument('-d', '--debug', action='store_true',
                            help='''Set the dash (flask) debug flag.''')
        return parser

    def run(self, args, tool):
        from ..gui.command_line import driver
        driver(args)


PluginClass = GUICommand



# Copyright (c) 2016  Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.

from .McsSubcommandABC import McsSubcommandABC

def driver(args, tool):
    """
    Start ipyparallel engines.
    """
    from pygcam.config import getParam
    from ..master import startEngines, templatePath

    scheduler = getParam('IPP.Scheduler')
    batchTemplate = templatePath(scheduler, args.profile, args.clusterId, 'engine')
    startEngines(args.numTrials, batchTemplate)


class EngineCommand(McsSubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Starts additional worker engines on a running cluster.'''}
        super(EngineCommand, self).__init__('engine', subparsers, kwargs)

    def addArgs(self, parser):
        from pygcam.config import getParam #, getParamAsInt

        # defaultQueue     = getParam('IPP.Queue')
        defaultProfile   = getParam('IPP.Profile')
        defaultClusterId = getParam('IPP.ClusterId')
        defaultWorkDir   = getParam('IPP.WorkDir')
        defaultNumTrials = 1

        parser.add_argument('-c', '--clusterId', type=str, default=defaultClusterId,
                            help='''A string to identify this cluster. Default is the
                            value of config var IPP.ClusterId, currently
                            "%s".''' % defaultClusterId)

        parser.add_argument('-n', '--numTrials', type=int, default=defaultNumTrials,
                            help='''The number of additional trials to create engines for.
                            Default is %d''' % defaultNumTrials)

        parser.add_argument('-o', '--otherArgs', type=str, default='',
                            help='Command line arguments to append to the ipengine command.')

        parser.add_argument('-p', '--profile', type=str, default=defaultProfile,
                            help='''The name of the ipython profile to use. Default is
                            the value of config var IPP.Profile, currently
                            "%s".''' % defaultProfile)

        # parser.add_argument('-q', '--queue', type=str, default=defaultQueue,
        #                     help='''The queue or partition on which to create the controller
        #                     and engines. Overrides config var IPP.Queue, currently
        #                     "%s".''' % defaultQueue)

        parser.add_argument('-w', '--workDir', type=str, default=defaultWorkDir,
                            help='''Where to run the ipengine command. Overrides the
                            value of config var IPP.WorkDir, currently '%s'.''' % defaultWorkDir)

        return parser   # for auto-doc generation


    def run(self, args, tool):
        driver(args, tool)



# Created on Mar 9, 2012
#
# @author: Richard Plevin
# @author: Sam Fendell
# @author: Ryan Jones
#
# Copyright (c) 2012-2015. The Regents of the University of California (Regents)
# and Richard Plevin. See the file COPYRIGHT.txt for details.
'''
This module includes contributions by Sam Fendell and Ryan Jones.
'''
import six
from collections import Iterable
from contextlib import contextmanager
from datetime import datetime
from operator import itemgetter
import sys

from sqlalchemy import create_engine, Table, Column, String, Float, text, MetaData, event
from sqlalchemy.engine import Engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import sessionmaker, load_only
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.pool import QueuePool

from pygcam.config import getSection, getParam, getParamAsBoolean
from pygcam.log import getLogger

from . import util as U
from .constants import RegionMap
from .error import PygcamMcsUserError, PygcamMcsSystemError
from .schema import (ORMBase, Run, Sim, Input, Output, InValue, OutValue, Experiment,
                     Program, Code, Region, TimeSeries)

_logger = getLogger(__name__)

def usingSqlite():
    '''
    Return True if the DbURL indicates that we're using Sqlite, else return False
    '''
    url = getParam('MCS.DbURL')
    return url.lower().startswith('sqlite')

def usingPostgres():
    '''
    Return True if the DbURL indicates that we're using Postgres, else return False
    '''
    url = getParam('MCS.DbURL')
    return url.lower().startswith('postgres')


@event.listens_for(Engine, "connect")
def sqlite_FK_pragma(dbapi_connection, connection_record):
    '''Turn on foreign key support in sqlite'''
    if usingSqlite():
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON")
        cursor.close()

    # TODO: might be useful:
    # pd.read_sql_table('data', engine, columns=['Col_1', 'Col_2'])
    # pd.read_sql_table('data', engine, index_col='id')
    # pd.read_sql_table('data', engine, parse_dates=['Date'])
    # pd.read_sql_table('data', engine, parse_dates={'Date': '%Y-%m-%d'})


RegionAliases = {
    'all regions':   'global',
    'rest of world': 'multiple',
    'row':           'multiple'
}

# The name of the program as stored in the "program" table
GCAM_PROGRAM = 'gcam'

# Constants to avoid misspelling errors
DFLT_PROGRAM  = 'default'

# Status strings for Run table
RUN_NEW       = 'new'
RUN_QUEUED    = 'queued'
RUN_RUNNING   = 'running'
RUN_SUCCEEDED = 'succeeded'
RUN_FAILED    = 'failed'
RUN_KILLED    = 'killed'
RUN_ABORTED   = 'aborted'
RUN_ALARMED   = 'alarmed'
RUN_UNSOLVED  = 'unsolved'
RUN_GCAMERROR = 'gcamerror'     # any other GCAM runtime error
ENG_TERMINATE = 'terminate'

RUN_FAILURES  = [RUN_FAILED, RUN_KILLED, RUN_ABORTED, RUN_ALARMED, RUN_UNSOLVED,
                 RUN_GCAMERROR]
RUN_STATUSES  = [RUN_NEW, RUN_QUEUED, RUN_RUNNING, RUN_SUCCEEDED] + RUN_FAILURES


# TBD: maybe drop this and store it from Context instead
def beforeSavingRun(_mapper, _connection, run):
    '''
    Before inserting/updating a Run instance, set numerical status and
    timestamps according to the status string.
    '''
    if run.status in (RUN_NEW, RUN_QUEUED):
        run.queueTime = datetime.now()
        run.startTime = None
        run.endTime   = None
        run.duration  = None
    elif run.status == RUN_RUNNING:
        run.startTime = datetime.now()
        run.endTime   = None
        run.duration  = None
    elif run.startTime:
        run.endTime   = datetime.now()
        delta         = run.endTime - run.startTime
        run.duration  = delta.seconds / 60


# Associate a listener function with Run, to execute before inserts and updates
event.listen(Run, 'before_insert', beforeSavingRun)
event.listen(Run, 'before_update', beforeSavingRun)

def parseSqlScript(filename=None, text=None):
    '''
    Parse a SQL script into semi-colon-terminated statements.

    :param filename: (str) the path of the file of SQL statements
    :param text: (str) optionally pass the contents of the file
      rather than the filename
    :return: (list of str) the individual statements
    '''
    from StringIO import StringIO

    if not (filename or text):
        raise PygcamMcsSystemError('Called parseSqlScript with neither filename nor text')

    if text:
        lines = StringIO(text).readlines()
    else:
        with open(filename) as f:
            lines = f.readlines()

    statements = []
    buffer = ''

    for line in lines:
        line = line.strip()
        if not line or line.startswith('--'):
            continue
        buffer += line
        if buffer[-1] == ';':
            statements.append(buffer)
            buffer = ''
        else:
            buffer += ' '

    return statements

class CoreDatabase(object):

    def __init__(self):
        from sqlalchemy.orm import scoped_session
        factory = sessionmaker() # N.B. sessionmaker returns a class object
        self.Session = scoped_session(factory)

        self.url     = None
        self.engine  = None
        self.appId   = None

    def endSession(self, session):
        '''
        Helper method to handle thread-scoped session objects for use with ipyparallel

        :param session: the open session
        :return: none
        '''
        try:
            self.Session.remove()
        except Exception as e:
            _logger.debug("Can't remove Session: %s", e)
            session.close()

    @contextmanager
    def sessionScope(self, withRetry=True):
        """
        Provide a transactional scope around a series of operations.
        Usage:
           with sessionScope() as session:
        """
        session = self.Session()
        session.expire_on_commit = False

        try:
            yield session
            if withRetry:
                self.commitWithRetry(session)
            else:
                session.commit()
        except:
            session.rollback()
            raise
        finally:
            self.endSession(session)

    # create_engine(*args, **kwargs)
    # The string form of the URL is dialect+driver://user:password@host/dbname[?key=value..],
    # where dialect is a database name such as mysql, oracle, postgresql, etc., and driver
    # the name of a DBAPI, such as psycopg2, pyodbc, cx_oracle, etc.
    # psycopg2: http://www.stickpeople.com/projects/python/win-psycopg/
    #
    # postgres|mysql: engine = create_engine('postgresql://scott:tiger@localhost/mydatabase')
    #
    # sqlite:   engine = create_engine('sqlite:///foo.db') -- if relative pathname, or
    #           engine = create_engine('sqlite:////absolute/path/to/foo.db') if abs pathname.
    #
    # To add a user and database in postgres:
    #    createuser -P mcsuser      # -P => prompt for password
    #    createdb -O mcsuser mcs    # -O => make mcsuser owner of database mcs
    #
    def startDb(self, checkInit=True):
        '''
        Links session to the database file identified in the config file as 'dbfile'.
        This needs to be called before any database operations can occur. It is called
        in getDatabase() when a new database instance is created.
        '''
        url  = getParam('MCS.DbURL')
        echo = getParamAsBoolean('MCS.EchoSQL')

        _logger.info('Starting DB: %s' % url)

        self.createDatabase()

        connect_args = {'connect_timeout': 15} if usingPostgres() else {}

        self.engine = engine = create_engine(url, echo=echo, connect_args=connect_args,
                                             # poolclass=QueuePool
                                             #, pool_pre_ping=True
                                            )
        self.Session.configure(bind=engine)

        self.url = url

        if checkInit:
            # Load metadata from the existing database, not from the ORMBase,
            # to see if the "run" table exists. If not, initialize the DB.
            # We don't do this if calling from Runner.py, which requires that
            # the database be set up already.
            meta = MetaData(bind=engine, reflect=True)
            if 'run' not in meta.tables:
                self.initDb()


    def initDb(self, args=None):
        '''
        Initialize the database, including loading required inserts.
        '''
        _logger.info('Initializing DB: %s' % self.url)

        meta = ORMBase.metadata     # accesses declared tables
        meta.bind = self.engine
        meta.reflect()
        meta.drop_all()

        session = self.Session()
        meta.create_all()
        session.commit()

        if args and args.empty:
            return

        # Deprecated?
        _logger.debug('Adding standard codes')
        # Add standard app status codes
        session.add(Code(codeName=RUN_QUEUED,    description='Trial queued'))
        session.add(Code(codeName=RUN_RUNNING,   description='Trial running'))
        session.add(Code(codeName=RUN_SUCCEEDED, description='Trial succeeded'))
        session.add(Code(codeName=RUN_FAILED,    description='Trial failed'))
        session.add(Code(codeName=RUN_ABORTED,   description='Runtime error'))
        session.add(Code(codeName=RUN_KILLED,    description='System timeout'))
        session.add(Code(codeName=RUN_ALARMED,   description='Runner timeout.'))

        #session.add(Program(name=DFLT_PROGRAM, description='Program name used when none is specified'))
        _logger.debug('Committing standard codes')
        session.commit()

        initialData = [
            ['Program',    [{'name': GCAM_PROGRAM,
                             'description' : 'The GCAM executable program'}]],
        ]

        _logger.debug('Adding initial data')
        for key, value in initialData:
            # If no module is specified with the class name, it is
            # assumed to be in this module, otherwise everything
            # up to the last '.' is treated as the module name
            items     = key.rsplit('.', 1)
            modName   = items[0] if len(items) == 2 else __name__
            className = items[1] if len(items) == 2 else items[0]

            table = className.lower()
            if table in ORMBase.metadata.tables:
                module    = sys.modules[modName]
                dataClass = getattr(module, className)
                if not dataClass:
                    raise PygcamMcsSystemError("Table class %s not found in module %s" % (className, modName))

                for row in value:
                    session.add(dataClass(**row))
                    _logger.debug('committing row')
                    session.commit()                # commit each row so each can refer to prior rows
            else:
                raise KeyError(table)


    def createDatabase(self):
        '''
        Ensure that the database directory (in the case of sqlite3) or the database is available.
        '''
        if usingSqlite():
            # Make sure required directory exists
            dbDir = getParam('MCS.RunDbDir')
            U.mkdirs(dbDir)
            return

        if usingPostgres() and getParam('MCS.Postgres.CreateDbExe'):
            import subprocess, shlex, re
            from .error import PygcamMcsSystemError

            # Make sure required database exists
            dbName   = getSection()
            createdb = getParam('MCS.Postgres.CreateDbExe')
            argStr   = getParam('MCS.Postgres.CreateDbArgs')
            command  = "%s %s" % (createdb, argStr)
            _logger.debug("Trying command: %s" % command)
            args = shlex.split(command)

            proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

            if not proc:
                raise PygcamMcsSystemError("Could not run: %s" % command)

            status = proc.wait()      # wait for process to complete
            if status == 0:
                _logger.debug("Created postgres database '%s'", dbName)
            else:
                output = list(proc.stdout)
                existsMsg = 'database "%s" already exists' % dbName
                dbExists  = any(map(lambda line: re.search(existsMsg, line), output))

                if not dbExists:
                    raise PygcamMcsSystemError("%s failed: %s" % (command, dbExists))     # fail if unexpected error occurs

                _logger.debug("Postgres database '%s' already exists", dbName)

    def commitWithRetry(self, session, maxTries=10, maxSleep=2.0):
        # N.B. With master/worker architecture, this should no longer be necessary, but
        # there are still occasional failures due to inability to acquire file lock.
        # import sqlite3
        import random
        import time

        tries = 0

        done = False
        while not done:
            try:
                session.commit()
                done = True

            except Exception as e:
            # except sqlite3.OperationalError as e:
                _logger.debug('sqlite3 operational error: %s', e)

                if tries >= maxTries:
                    raise PygcamMcsSystemError("Failed to acquire database lock")

                delay = random.random() * maxSleep    # sleep for a random number of seconds up to maxSleep
                _logger.warn("Database locked (retry %d); sleeping %.1f sec" % (tries, delay))
                time.sleep(delay)
                tries += 1

            # except Exception as e:
            #     raise PygcamMcsSystemError("commitWithRetry error: %s" % e)


    def execute(self, sql):
        'Execute the given SQL string'
        _logger.debug('Executing SQL: %s' % sql)
        with self.sessionScope() as session:
            session.execute(sql)

    def executeScript(self, filename=None, text=None):
        if not (filename or text):
            raise PygcamMcsSystemError('Called executeScript with neither filename nor text')

        lines = parseSqlScript(filename=filename, text=text)

        with self.sessionScope() as session:
            for sql in lines:
                _logger.debug('Executing script SQL: %s' % sql)
                session.execute(sql)

    def dropAll(self):
        '''
        Drop all objects in the database, even those not defined in SqlAlchemy.
        '''
        with self.sessionScope() as session:
            engine  = session.get_bind()

            meta = MetaData(bind=engine)
            meta.reflect()

            if usingPostgres():
                self.dropAllPostgres(meta)
            else:
                meta.drop_all()

    def dropAllPostgres(self, meta):
        """
        Drop all postgres tables and sequences from schema 'public'
        """
        for type in ('table', 'sequence'):
            sql = "select {type}_name from information_schema.{type}s where {type}_schema='public'".format(type=type)
            names = [name for (name, ) in meta.execute(text(sql))]
            for name in names:
                try:
                    meta.execute(text("DROP %s %s CASCADE" % (type, name)))
                except SQLAlchemyError as e:
                    print e

    def getTable(self, tableClass, orderBy=None):
        '''
        Get all the contents of a table, optionally ordered by the given column.

        :param tableClass: (SqlAlchemy table class) the table to query
        :param orderBy: (SqlAlchemy table.col spec) optional column to order by
        :return: (list of elements of tableClass) the rows of the table
        '''
        with self.sessionScope() as session:
            query = session.query(tableClass)
            if orderBy:
                query = query.order_by(orderBy)

            rows = query.all()
            return rows

    def setSqlEcho(self, value=True):
        with self.sessionScope() as session:
            engine = session.get_bind()
            engine.echo = value

    def addColumns(self, tableClass, columns):
        '''
        Adds a new column or columns to an existing table, emitting an
        ALTER TABLE statement and updating the metadata.
        :param tableClass: the class defining the table to alter
        :param column: a Column instance describing the column to add
        :return: none
        '''
        if not isinstance(columns, Iterable):
            columns = [columns]

        with self.sessionScope() as session:
            engine  = session.get_bind()
            table   = tableClass.__table__
            tableName  = table.description
            for column in columns:
                table.append_column(column)                     # add it to the metadata
                setattr(tableClass, column.name, column)        # add attribute to class, which maps it to the column
                columnName = column.name                        # column.compile(dialect=engine.dialect)
                columnType = column.type.compile(engine.dialect)
                engine.execute('ALTER TABLE %s ADD COLUMN "%s" %s' % (tableName, columnName, columnType))

    def createOutput(self, name, description=None, program=DFLT_PROGRAM, session=None):
        '''
        Create an Output record with the given arguments unless the record
        exists for this name and program. Uses caller's session if provided. If no
        session is provided, the row is committed and the new outputId is returned.
        If the caller passes a session, None is returned unless the object was found,
        in which case the outputId is returned.
        '''
        sess      = session or self.Session()
        programId = self.getProgramId(program)
        outputId  = sess.query(Output.outputId).filter_by(programId=programId, name=name).scalar()

        if outputId is None:
            output = Output(name=name, programId=programId, description=description)
            sess.add(output)

            if not session:
                sess.commit()
                self.endSession(sess)
                outputId = output.outputId

        return outputId

    def getOutputIds(self, nameList):
        # cache on first call
        if not self.outputIds:
            with self.sessionScope() as session:
                rows = session.query(Output.name, Output.outputId).all()
                self.outputIds = dict(rows)

        # lookup ids in cache
        ids = [self.outputIds[name] for name in nameList]
        return ids
        # return zip(*ids)[0] if ids else []

    def getOutputs(self):
        rows = self.getTable(Output)
        return map(lambda obj: obj.name, rows)

    def getOutputsWithValues(self, simId, scenario):
        with self.sessionScope() as session:
            query = session.query(Output.name).\
                join(OutValue).join(Run).filter_by(simId=simId).\
                join(Experiment).filter_by(expName=scenario). \
                distinct(Output.name)
            rows = query.all()
            return map(lambda row: row[0], rows)

    #
    # Very much like setAttrVal. So much so that perhaps the Result table can be
    # eliminated in favor of using the generic attribute/value system?
    #
    def setOutValue(self, runId, paramName, value, program=GCAM_PROGRAM, session=None):
        '''
        Set the given named output parameter to the given numeric value. Overwrite a
        previous value for this runId and attribute, if found, otherwise create a new value
        record. If session is not provided, one is allocated, and the transaction is
        committed. If a session is provided, the caller is responsible for calling commit.
        '''
        #_logger.debug('setOutValue(%s, %s, %s, session=%s', runId, paramName, value, session)
        sess = session or self.Session()

        outputId = sess.query(Output.outputId).filter_by(name=paramName).join(Program).filter_by(name=program).scalar()
        if not outputId:
            raise PygcamMcsSystemError("%s output %s was not found in the Output table" % (program, paramName))

        results = sess.query(OutValue).filter_by(runId=runId, outputId=outputId).all()

        # If previous value is found, overwrite it; otherwise create a new one
        if results:
            #_logger.debug("setOutValue: updating value for outputId=%d" % outputId)
            #result = resultQuery.one()
            result = results[0]
            result.value = value
        else:
            #_logger.debug("setOutValue: adding value for outputId=%d" % outputId)
            sess.add(OutValue(runId=runId, outputId=outputId, value=value))

        if session is None:
            self.commitWithRetry(sess)
            self.endSession(sess)

    def getOutValues(self, simId, expName, outputName, limit=None):
        '''
        Return a pandas DataFrame with columns trialNum and name outputName,
        for the given sim, exp, and output variable.
        '''
        from pandas import DataFrame

        session = self.Session()

        limit = sys.maxint if limit is None or limit <= 0 else limit

        # This is essentially this query, but with "JOIN xx ON" syntax generated:
        #   select r.trialNum, v.value from run r, outvalue v, experiment e, output o
        #   where e.scenario='test exp' and r.expid=e.expid and r.simid=1 and
        #         o.name='p1' and o.outputid=v.outputid;
        query = session.query(Run.trialNum).add_columns(OutValue.value).filter_by(simId=simId).\
        join(Experiment).filter_by(expName=expName).join(OutValue).join(Output).filter_by(name=outputName).\
        order_by(Run.trialNum).limit(limit)

        #print "getOutValues query: %s" % str(query.statement.compile())

        rslt = query.all()
        self.endSession(session)

        if not rslt:
            return None

        resultDF = DataFrame.from_records(rslt, columns=['trialNum', outputName], index='trialNum')
        return resultDF

    def deleteRunResults(self, runId, outputIds=None, session=None):
        sess = session or self.Session()

        query = sess.query(OutValue).filter_by(runId=runId)
        if outputIds:
            query = query.filter(OutValue.outputId.in_(outputIds))

        #query.delete(synchronize_session='fetch')
        query.delete(synchronize_session=False)

        if session is None:
            self.commitWithRetry(sess)
            self.endSession(sess)

    # def queryToDataFrame(self, query):  # TBD: Not used anywhere yet...
    #     from pandas import DataFrame    # lazy import
    #
    #     session = self.Session()
    #     result = session.execute(query)
    #     columnNames = result.keys()
    #     values = result.fetchall()
    #     self.endSession(session)
    #     return DataFrame(values, columns=columnNames)

    def getParameterValues(self, simId, program='gcam', asDataFrame=False):
        from pandas import DataFrame    # lazy import
        session = self.Session()

        query = session.query(InValue.row, InValue.col, InValue.value, InValue.trialNum, Input.paramName).\
                 filter(InValue.simId == simId).join(Input).join(Program).filter(Program.name == program).order_by(InValue.trialNum)

        rslt = query.all()
        cols = map(lambda d: d['name'], query.column_descriptions) if rslt else None
        self.endSession(session)

        if not rslt:
            return None

        if not asDataFrame:
            return rslt

        resultDF = DataFrame.from_records(rslt, columns=cols, index='trialNum')
        return resultDF

    def getParameterValues2(self, simId):
        from pandas import DataFrame    # lazy import
        session = self.Session()

        query = session.query(InValue.trialNum, InValue.value, Input.paramName).\
                 filter(InValue.simId == simId).join(Input).order_by(InValue.trialNum)

        rslt = query.all()
        self.endSession(session)

        if not rslt:
            return None

        cols = ['trialNum', 'value', 'paramName']
        resultDF = DataFrame.from_records(rslt, columns=cols)
        resultDF = resultDF.pivot(index='trialNum', columns='paramName', values='value')
        return resultDF

    def getParameters(self):
        session = self.Session()
        query = session.query(Input.paramName, InValue.row, InValue.col).\
            filter(Input.inputId == InValue.inputId).\
            distinct(Input.paramName, InValue.row, InValue.col)

        rslt = query.all()
        self.endSession(session)
        return rslt

    def getInputs(self):
        rows = self.getTable(Input, orderBy=Input.paramName)
        return map(lambda row: row.paramName, rows)

    def scenariosWithResults(self, simId):
        # Definition of view 'result':
        # select o."runId", r."simId", r."expId", r."trialNum", e."expName", op.name, o.value
        # from outvalue o, output op, run r, experiment e
        # where e."expId" = r."expId" and o."runId" = r."runId" and o."outputId" = op."outputId"
        try:
            with self.sessionScope() as session:
                query = session.query(Experiment.expName).join(Run).filter_by(simId=simId).\
                    join(OutValue).distinct(Experiment.expName)
                rows = query.all()
                names = map(lambda row: row[0], rows)
        except Exception as e:
            _logger.error("scenariosWithResults failed: %s", e)
            names = []

        _logger.debug("scenariosWithResults returning %s", names)
        return names


    def createRun(self, simId, trialNum, expName=None, expId=None, status=RUN_NEW, session=None):
        """
        Create an entry for a single model run, initially in "new" state
        """
        assert (expName or expId), "Database createRun called with neither expName nor expId"

        sess = session or self.Session()
        if expId is None:
            exp = sess.query(Experiment.expId).filter_by(expName=expName).one()
            expId = exp.expId

        # if prior run record exists for this {simId, trialNum, expId} tuple, delete it
        with session.no_autoflush:
            sess.query(Run).filter_by(simId=simId, trialNum=trialNum, expId=expId).delete()

        run = Run(simId=simId, trialNum=trialNum, expId=expId, status=status, jobNum=None)
        sess.add(run)

        if not session:     # if we created the session locally, commit; else call must do so
            self.commitWithRetry(sess)
            self.endSession(sess)

        return run

    def getSim(self, simId):
        with self.sessionScope() as session:
            sim = session.query(Sim).filter_by(simId=simId).scalar()

            return sim    # N.B. scalar() returns None if no rows are found

    def getSims(self):
        rows = self.getTable(Sim, orderBy=Sim.simId)
        return rows

    def getRun(self, simId, trialNum, expName):
        with self.sessionScope() as session:
            run = session.query(Run).filter_by(simId=simId, trialNum=trialNum).\
                    join(Experiment).filter_by(expName=expName).scalar()

            return run

    def getRunByRunId(self, runId):
        with self.sessionScope() as session:
            run = session.query(Run).filter_by(runId=runId).scalar()
            return run

    def getRunFromContext(self, context):
        run = self.getRun(context.simId, context.trialNum, context.scenario)
        #_logger.debug("getRunIdFromContext returning runId %s", run.runId if run else None)
        return run

    def setRunStatus(self, runId, status, session=None):
        '''
        Set the runStatus to the value for the given string and
        optionally set the job number.'''
        sess = session or self.Session()
        try:
            run = sess.query(Run).filter_by(runId=runId).one()
            if run.status == status:
                return # nothing to do here

            run.status = status    # insert/update listener sets status code and timestamps

            if not session:
                self.commitWithRetry(sess)
            return

        except NoResultFound:
            _logger.warn("db.setRunStatus failed to find record for runId %d", runId)
            return

        finally:
            if not session:
                self.endSession(sess)

    def getRunsWithStatus(self, simId, expList, statusList):
        from operator import itemgetter         # lazy import

        # Allow expList and statusList to be a single string,
        # which we convert to lists
        if isinstance(expList, six.string_types):
            expList = [expList]

        if isinstance(statusList, six.string_types):
            statusList = [statusList]

        session = self.Session()
        query = session.query(Run.trialNum).filter_by(simId=simId)

        if expList:
            query = query.join(Experiment).filter(Run.status.in_(statusList), Experiment.expName.in_(expList))

        rslt = query.order_by(Run.trialNum).all()
        self.endSession(session)

        if rslt:
            rslt = map(itemgetter(0), rslt) # collapse list of singleton tuples into a single list

        #_logger.debug("for simid=%d, expList=%s, status=%s, rslt=%s" % (simId, expList, status, rslt))
        return rslt

    def getRunsByStatus(self, simId, scenario, statusList, groupName=None, projectName=None):
        '''
        By default, returns tuples of (runId, trialNum) for the given scenario that have
        any of the statuses in statusList (which can be a single status string or a list
        of strings.) If groupName or projectName are not None, results are converted to
        a list of Context instances.
        '''
        from .context import Context

        if isinstance(statusList, six.string_types):
            statusList = [statusList]

        if len(statusList) == 0:
            return []

        with self.sessionScope() as session:

            # expId = self.getExpId(scenario, session=session)
            # query = session.query(Run.runId, Run.trialNum).filter_by(simId=simId, expId=expId).filter(Run.status.in_(statusList))

            # Return all data required to create Context (except projectName and groupName)
            query = session.query(Run.runId, Run.simId, Run.trialNum, Run.status).filter_by(simId=simId).filter(Run.status.in_(statusList))
            query = query.add_columns(Experiment.expName, Experiment.parent).join(Experiment).filter_by(expName=scenario)

            rslt = query.order_by(Run.trialNum).all()

        if groupName or projectName:
            rslt = map(lambda r: Context(runId=r[0], simId=r[1], trialNum=r[2], status=r[3], scenario=r[4],
                                         baseline=r[5], groupName=groupName, projectName=projectName),
                       rslt)
        return rslt

    def createSim(self, trials, description, simId=None):
        '''
        Creates a new simulation with the given number of trials and description
        '''
        with self.sessionScope() as session:
            if simId is None:
                newSim = Sim(trials=trials, description=description)
            else:
                session.query(Sim).filter_by(simId=simId).delete()
                newSim = Sim(trials=trials, description=description, simId=simId)

            session.add(newSim)

        return newSim.simId

    def updateSimTrials(self, simId, trials):
        with self.sessionScope() as session:
           sim = session.query(Sim).filter_by(simId=simId).one()
           sim.trials = trials

    def getTrialCount(self, simId):
        with self.sessionScope() as session:
            trialCount = session.query(Sim.trials).filter_by(simId=simId).scalar()
            return trialCount

    def getMissingTrials(self, simId, scenario):
        """
        Return a list of trial numbers that are not present in the database
        for the given simId and scenario.

        :param simId: (int) simulation ID
        :param scenario: (str) scenario name
        :return: (list of int) trial numbers of missing trials
        """
        count = self.getTrialCount(simId)
        possible = set(xrange(count))

        df = self.getRunInfo(simId, scenario, includeSucceededRuns=True, asDataFrame=True)
        present = set() if df is None else set(df.trialNum)

        missing = possible - present
        return sorted(missing)

    def getRunInfo(self, simId, scenario, includeSucceededRuns=False, asDataFrame=False):
        """
        Return info for runs for the given simId and scenario, as a list of tuples
        or as a pandas.DataFrame.

        :param simId: (int) simulation ID
        :param scenario: (str) scenario name
        :param includeSucceededRuns: (bool) if True, runs of status 'succeeded' are
           included; by default they are not.
        :param asDataFrame: (bool) whether to return the result as a list of tuples
           (the default) or as a DataFrame
        :return:
        """
        from pandas import DataFrame

        with self.sessionScope() as session:
            q = session.query(Run.runId, Run.simId, Run.trialNum, Run.status).\
                    filter_by(simId=simId).join(Experiment).filter_by(expName=scenario)

            if not includeSucceededRuns:
                q = q.filter(Run.status != 'succeeded')

            rows = q.all()
            if asDataFrame:
                if rows:
                    cols = map(lambda d: d['name'], q.column_descriptions)
                    df = DataFrame.from_records(rows, columns=cols, index='runId')
                    return df
                return None

            return rows

    def createExp(self, name, description=None):
        '''
        Insert a row for the given experiment
        '''
        with self.sessionScope() as session:
            exp = Experiment(expName=name, description=description)
            session.add(exp)
            return exp.expId

    def getExpId(self, expName, session=None):
        exp = self.getExp(expName, session)
        return exp.expId

    def getExp(self, expName, session=None, raiseError=True):
        sess = session or self.Session()

        try:
            exp = sess.query(Experiment).filter_by(expName=expName).one()

        except NoResultFound:
            msg = "The experiment '%s' is not defined" % expName
            if raiseError:
                _logger.fatal(msg)
                raise PygcamMcsUserError(msg)
            else:
                _logger.info(msg)
                exp = None

        finally:
            if session:
                self.endSession(session)

        return exp

    def addExperiments(self, scenarioNames, baseline, filename):
        from .error import PygcamMcsSystemError

        desc = 'Added from ' + filename

        for name in scenarioNames:
            parent = None if name == baseline else baseline
            try:
                self.createExp(name, description=desc, parent=parent)

            except Exception as e:
                raise PygcamMcsSystemError("Failed to create experiment: %s" % e)

    def getProgramId(self, program):
        with self.sessionScope() as session:
            programId = session.query(Program.programId).filter_by(name=program).scalar()
            return programId


class GcamDatabase(CoreDatabase):
    _yearColsAdded = False
    _expColsAdded = False

    instance = None     # singleton class

    def __init__(self):
        super(GcamDatabase, self).__init__()
        self.paramIds = {}                   # parameter IDs by name
        self.outputIds = None                # output IDs by name
        self.canonicalRegionMap = {}

        # Cache these to avoid database access in saveResults loop
        for regionName, regionId in RegionMap.items():
            canonName = canonicalizeRegion(regionName)
            self.canonicalRegionMap[canonName] = regionId

    @classmethod
    def getDatabase(cls, checkInit=True):
        if cls.instance is None:
            cls.instance = GcamDatabase()
            cls.instance.startDb(checkInit=checkInit)

        return cls.instance

    @classmethod
    def close(cls):
        if cls.instance:
            cls.instance.engine.dispose()
        cls.instance = None

    def initDb(self, args=None):
        'Add GCAM-specific tables to the coremcs database'
        super(GcamDatabase, self).initDb(args=args)

        self.addYearCols()
        self.addExpCols()

        if args and args.empty:
            return

        self.addRegions(RegionMap)

    def startDb(self, checkInit=True):
        super(GcamDatabase, self).startDb(checkInit=checkInit)
        self.addYearCols(alterTable=False)
        self.addExpCols(alterTable=False)

    def createExp(self, name, parent=None, description=None):
        '''
        Insert a row for the given experiment. Replaces superclass method
        to add 'parent' argument. Also, if it fails, updates existing row.
        '''
        from sqlalchemy.exc import IntegrityError

        session = self.Session()
        exp = Experiment(expId=None, expName=name, description=description)
        exp.parent = parent # not in Experiment's __init__ signature

        try:
            session.add(exp)
            session.commit()
            expId = exp.expId

        except IntegrityError:
            session.rollback()
            expId = self.updateExp(name, description=description, parent=parent, session=session)

        finally:
            self.endSession(session)

        return expId

    def updateExp(self, name, parent=None, description=None, session=None):
        sess = session or self.Session()
        exp = self.getExp(name, session=sess)
        exp.description = description
        exp.parent = parent
        sess.commit()

        if not session:
            self.endSession(sess)

        return exp.expId

    def getExp(self, expName, session=None, raiseError=True):
        sess = session or self.Session()

        try:
            exp = sess.query(Experiment).filter_by(expName=expName).one()

        except NoResultFound:
            msg = "The experiment '%s' is not defined" % expName
            if raiseError:
                raise PygcamMcsUserError(msg)
            else:
                _logger.info(msg)
                exp = None

        finally:
            if not session:
                self.endSession(sess)

        return exp

    def getExps(self):
        rows = self.getTable(Experiment, orderBy=Experiment.expName)
        return rows

    # TBD: generalize and add to CoreDatabase since often modified? Or just add to base schema.
    def addExpCols(self, alterTable=True):
        '''
        Add required columns to the Experiment table.
        '''
        if self._expColsAdded:
            return

        session = self.Session()
        engine = session.get_bind()
        meta = ORMBase.metadata
        meta.bind = engine
        meta.reflect()
        expTable = Experiment.__table__

        cols = [('parent', String)]

        for (colName, colType) in cols:
            if colName not in expTable.columns:
                column = Column(colName, colType)
                if alterTable:
                    self.addColumns(Experiment, column)
                else:
                    setattr(Experiment, column.name, column)        # just add the mapping

        self._expColsAdded = True
        self.endSession(session)

    @staticmethod
    def yearCols():
        from .util import activeYears, YEAR_COL_PREFIX

        # Create the time series table with years (as columns) specified in the config file.
        years = activeYears()
        cols = map(lambda y: YEAR_COL_PREFIX + y, years)
        return cols

    def addYearCols(self, alterTable=True):
        '''
        Define year columns (y1990, y2005, y2010, etc.) dynamically. If alterTable
        is True, the SQL table is altered to add the column; otherwise the column
        is just mapped to an attribute of the TimeSeries class.
        '''
        if self._yearColsAdded:
            return

        with self.sessionScope() as session:
            engine = session.get_bind()
            meta = ORMBase.metadata
            meta.bind = engine
            meta.reflect()

            colNames = self.yearCols()

            timeSeries = Table('timeseries', meta, autoload=True, autoload_with=engine)

            # Add columns for all the years used in this analysis
            for colName in colNames:
                if colName not in timeSeries.columns:
                    column = Column(colName, Float)
                    if alterTable:
                        self.addColumns(TimeSeries, column)
                    else:
                        setattr(TimeSeries, column.name, column)        # just add the mapping

            self._yearColsAdded = True

    def addRegions(self, regionMap):
        # TBD: read region map from file identified in config file, or use default values
        # For now, use default mapping
        with self.sessionScope() as session:
            for name, regId in regionMap.iteritems():
                self.addRegion(regId, name, session=session)

    def addRegion(self, regionId, name, session=None):
        sess = session or self.Session()
        obj = Region(regionId=regionId, displayName=name, canonName=canonicalizeRegion(name))
        sess.add(obj)

        if session:
            sess.commit()
            self.endSession(sess)

    def getRegionId(self, name):
        canonName = canonicalizeRegion(name)
        regionId = self.canonicalRegionMap[canonName]
        return regionId

    def getParamId(self, pname):
        return self.paramIds[pname]

    def createOutput(self, name, program=GCAM_PROGRAM, description=None, session=None):
        _logger.debug("createOutput(%s)", name)
        return super(GcamDatabase, self).createOutput(name, program=program, description=description, session=session)

    def saveParameterNames(self, tuples):
        '''
        Define parameter names in the database on the fly based on results of XPath queries.
        "tuples" is a list of (paramName, description) pairs.
        '''
        session = self.Session()
        programId = self.getProgramId(GCAM_PROGRAM)

        # TBD: The following code is subject to a race condition, but we don't expect multiple users to
        # TBD: generate simulations in the same model run dir simultaneously. If they do, this may break.
        # TBD: Could handle this with a lock...
        pnames = map(itemgetter(0), tuples)
        rows = session.query(Input).filter(Input.programId == programId, Input.paramName.in_(pnames)).all()
        found = map(lambda row: row.paramName, rows)

        descByName = dict(tuples)
        notFound = set(pnames) - set(found)

        # Construct list of tuples holding only the parameters that were
        # found, and whose description changed. These are updated automatically
        # when the session is committed.
        updTuples = []
        for row in rows:
            desc = descByName[row.paramName]
            if row.description != desc:
                row.description = desc
                updTuples.append(row)

        # Create a list of objects that need to be inserted, then add them all at once
        newInputs = map(lambda name: Input(programId=programId, paramName=name, description=descByName[name]), notFound)
        session.add_all(newInputs)

        # Insert new parameter descriptions for these
        session.commit()

        # Cache all parameter IDs for faster lookup
        rows = session.query(Input.inputId, Input.paramName).all()
        for row in rows:
            self.paramIds[row.paramName] = row.inputId

        self.endSession(session)

    def saveParameterValues(self, simId, tuples):
        '''
        Save the value of the given parameter in the database. Tuples are
        of the format: (trialNum, paramId, value, varNum)
        '''
        with self.sessionScope() as session:
            for trialNum, paramId, value, varNum in tuples:
                # We save varNum to distinguish among independent values for the same variable name.
                # The only purpose this serves is to ensure uniqueness, enforced by the database.
                inValue = InValue(inputId=paramId, simId=simId, trialNum=trialNum,
                                  value=value, row=0, col=varNum)
                session.add(inValue)

    def deleteRunResults(self, runId, outputIds=None, session=None):
        """
        Augment core method by deleting timeseries data, too.
        """
        # _logger.debug("deleteRunResults: deleting results for runId %d, outputIds=%s" % (runId, outputIds))
        sess = session or self.Session()
        super(GcamDatabase, self).deleteRunResults(runId, outputIds=outputIds, session=sess)

        query = sess.query(TimeSeries).filter_by(runId=runId)

        if outputIds:
            query = query.filter(TimeSeries.outputId.in_(outputIds))

        query.delete(synchronize_session='fetch')

        if session is None:
            self.commitWithRetry(sess)
            self.endSession(sess)

    def saveTimeSeries(self, runId, regionId, paramName, values, units=None, session=None):
        sess = session or self.Session()

        programId = self.getProgramId(GCAM_PROGRAM)

        # one() raises error if 0 or more than 1 row is found, otherwise returns a tuple.
        try:
            row = sess.query(Output).filter(Output.programId == programId, Output.name == paramName).one()
        except Exception:
            _logger.error("Can't find param %s for %s", paramName, GCAM_PROGRAM)
            raise

        outputId = row.outputId

        ts = TimeSeries(runId=runId, outputId=outputId, regionId=regionId, units=units)

        for name, value in values.iteritems():  # Set the values for "year" columns
            setattr(ts, name, value)

        sess.add(ts)

        if not session:
            sess.commit()
            self.endSession(sess)

    def getTimeSeries(self, simId, paramName, expList):
        '''
        Retrieve all timeseries rows for the given simId and paramName.

        :param simId: simulation ID
        :param paramName: name of output parameter
        :param expList: (list of str) the names of the experiments to select
           results for.
        :return: list of TimeSeries tuples or None
        '''
        cols = ['seriesId', 'runId', 'outputId', 'units'] + self.yearCols()

        with self.sessionScope() as session:
            query = session.query(TimeSeries, Experiment.expName).options(load_only(*cols)). \
                join(Run).filter_by(simId=simId).filter_by(status='succeeded'). \
                join(Experiment).filter(Experiment.expName.in_(expList)). \
                join(Output).filter_by(name=paramName)

            rslt = query.all()
            return rslt


# Single instance of the class. Use 'getDatabase' constructor
# to ensure that this instance is returned if already created.
_DbInstance = None

def getDatabase(checkInit=True):
    '''
    Return the instantiated CoreDatabase, or created one and return it.
    The optional dbClass argument is provided to facilitate subclassing.
    '''
    return GcamDatabase.getDatabase(checkInit=checkInit)

def dropTable(tableName, meta):
    if tableName in meta.tables:
        # Drop the table if it exists and remove it from the metadata
        table = meta.tables[tableName]
        table.drop()
        meta.remove(table)

def canonicalizeRegion(name):
    '''
    Return the canonical name for a region, normalizing the use of capitalization
    and underscores.

    :param name: a GCAM region name
    :return: region name in canonical format, i.e., lowercase with underscores
       changed to spaces. (The use of underscores is inconsistent and thus hard
       to remember, e.g., 'South America_Northern')
    '''
    name = name.lower()
    if name in RegionAliases:
        name = RegionAliases[name]

    return name.replace('_', ' ')



'''
.. gcamtool plugin for setting up / customizing GCAM project's XML files.

.. codeauthor:: Rich Plevin <rich@plevin.com>

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from ..subcommand import SubcommandABC

class SetupCommand(SubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help': '''Setup a scenario by creating modified XML input files.'''}

        super(SetupCommand, self).__init__('setup', subparsers, kwargs, group='project')

    def addArgs(self, parser):
        defaultYears = '2015-2100'

        group1 = parser.add_mutually_exclusive_group()  # --dynamicOnly, --staticOnly
        group2 = parser.add_mutually_exclusive_group()  # --modulePath, --moduleSpec, --setupXml

        parser.add_argument('-b', '--baseline',
                            help='''Identify the baseline the selected scenario is based on.
                                 Note: at least one of --baseline (-b) / --scenario (-s) must be used.''')

        # mutually exclusive with --staticOnly
        group1.add_argument('-d', '--dynamicOnly', action='store_true',
                            help='''Generate only dynamic XML for dyn-xml: don't create static XML.''')

        parser.add_argument('-f', '--forceCreate', action='store_true',
                            help='''Re-create the workspace, even if it already exists.''')

        parser.add_argument('-g', '--group',
                            help='The scenario group to process. Defaults to the group labeled default="1".')

        parser.add_argument('-G', '--srcGroupDir',
                            help='''A sub-directory under xmlsrc in which to find scenario dirs for this group.
                            Use this to consolidate static XML files shared by multiple scenario groups.
                            If --useGroupDir is specified, srcGroupDir defaults to the scenario group name.
                            Using --srcGroupDir implies --useGroupDir.''')

        # mutually exclusive with --moduleSpec and --setupXml
        group2.add_argument('-m', '--modulePath',
                            help='''The path to a scenario definition module. See -M flag for more info.''')

        # mutually exclusive with --modulePath and --setupXml
        group2.add_argument('-M', '--moduleSpec',
                            help='''The "dot spec" for the Python module holding the setup classes and
                            a function called 'scenarioMapper' or a dictionary called 'ClassMap' which map
                            scenario names to classes. If the function 'scenarioMapper' exists, it is
                            used. If not, the 'ClassMap' is used. Default is "{xmlsrc}/subdir/scenarios.py" (if
                            subdir is defined) or "{xmlsrc}/scenarios.py" (if subdir is undefined) under the
                            current ProjectRoot.''')

        parser.add_argument('-p', '--stop', type=int, metavar='period-or-year', dest='stopPeriod',
                            help='The number of the GCAM period or the year to stop after')

        parser.add_argument('-r', '--refWorkspace', default="",
                            help='''A reference workspace to use instead of the value of GCAM.RefWorkspace''')

        parser.add_argument('-R', '--resultsDir',
                            help='The parent directory holding the GCAM output workspaces')

        parser.add_argument('-s', '--scenario',
                            help='''Identify the scenario to run.
                            Note: at least one of --baseline (-b) / --scenario (-s) must be used.''')

        parser.add_argument('-S', '--subdir', default="",
                            help='A sub-directory to use instead of scenario name')

        # mutually exclusive with --moduleSpec and --modulePath
        group2.add_argument('--setupXml',
                            help='''An XML scenario definition file. Overrides configuration variable
                             GCAM.ScenarioSetupFile.''')

        # mutually exclusive with --dynamicOnly
        group1.add_argument('-T', '--staticOnly', action='store_true',
                            help='''Generate only static XML for local-xml: don't create dynamic XML.''')

        parser.add_argument('-u', '--useGroupDir', action='store_true',
                            help='Use the group name as a sub directory below xmlsrc, local-xml, and dyn-xml')

        parser.add_argument('-x', '--xmlSourceDir',
                            help='''The location of the xmlsrc directory.''')

        parser.add_argument('-X', '--xmlOutputRoot',
                            help='''The root directory into which to generate XML files.''')

        parser.add_argument('-w', '--workspace', # i.e., sandbox
                            help='''The pathname of the workspace to operate on.''')

        # Deprecated or pass to scenario?
        parser.add_argument('-y', '--years', default=defaultYears,
                            help='''Years to generate constraints for. Must be of the form
                            XXXX-YYYY. Default is "%s"''' % defaultYears)

        return parser   # for auto-doc generation


    def run(self, args, tool):
        import os
        from importlib import import_module

        from ..config import getParam
        from ..error import SetupException
        from ..log import getLogger
        from ..scenarioSetup import createSandbox
        from ..utils import loadModuleFromPath, pathjoin

        _logger = getLogger(__name__)

        scenario = args.scenario or args.baseline
        if not scenario:
            raise SetupException('At least one of --baseline (-b) / --scenario (-s) must be used.')

        projectDir = getParam('GCAM.ProjectDir')
        groupName = args.group if args.useGroupDir else ''
        srcGroupDir = args.srcGroupDir or groupName

        if args.workspace:
            workspace = args.workspace
        else:
            groupDir  = os.path.normpath(pathjoin(projectDir, groupName))
            workspace = pathjoin(groupDir, scenario)

        mcsMode = tool.getMcsMode()
        forceCreate = args.forceCreate or bool(mcsMode)

        if not mcsMode or mcsMode == 'trial':
            createSandbox(workspace, srcWorkspace=args.refWorkspace, forceCreate=forceCreate, mcsMode=mcsMode)

        xmlSourceDir = args.xmlSourceDir or getParam('GCAM.XmlSrc')

        # If a setup XML file is defined, use the defined (or default) XMLEditor subclass
        setupXml = args.setupXml or getParam('GCAM.ScenarioSetupFile')
        if setupXml:
            from ..xmlSetup import createXmlEditorSubclass
            _logger.debug('Setup using %s, mcsMode=%s', setupXml, mcsMode)
            scenClass = createXmlEditorSubclass(setupXml, mcsMode=mcsMode)

        else:
            # If neither is defined, we assume a custom scenarios.py file is used
            try:
                if args.moduleSpec:
                    module = import_module(args.moduleSpec, package=None)
                else:
                    modulePath = args.modulePath or pathjoin(xmlSourceDir, srcGroupDir, 'scenarios.py')
                    _logger.debug('Setup using %s', modulePath)
                    module = loadModuleFromPath(modulePath)

            except Exception as e:
                moduleName = args.moduleSpec or modulePath
                raise SetupException('Failed to load scenarioMapper or ClassMap from module %s: %s' % (moduleName, e))

            try:
                # First look for a function called scenarioMapper
                scenarioMapper = getattr(module, 'scenarioMapper', None)
                if scenarioMapper:
                    scenClass = scenarioMapper(scenario)

                else:
                    # Look for 'ClassMap' in the specified module
                    classMap  = getattr(module, 'ClassMap')
                    scenClass = classMap[scenario]

            except KeyError:
                raise SetupException('Failed to map scenario "%s" to a class in %s' % (scenario, module.__file__))

        subdir = args.subdir or scenario
        refWorkspace  = args.refWorkspace or getParam('GCAM.RefWorkspace')
        xmlOutputRoot = args.xmlOutputRoot or workspace

        # When called from gcammcs in 'trial' mode, we only run dynamic setup. When run
        # in 'gensim' mode, we do only static setup.
        args.dynamicOnly = args.dynamicOnly or mcsMode == 'trial'

        if mcsMode == 'gensim':
            args.dynamicOnly = False
            args.staticOnly  = True

        # TBD: Document that all setup classes must conform to this protocol
        obj = scenClass(args.baseline, args.scenario, xmlOutputRoot, xmlSourceDir,
                        refWorkspace, groupName, srcGroupDir, subdir)

        obj.mcsMode = mcsMode
        obj.setup(args)



#!/usr/bin/env python
'''
.. Copyright (c) 2016 Richard Plevin

   See the https://opensource.org/licenses/MIT for license details.
'''

# Dummy "gt" to tie together pygcam.mcs plugins to generate documentation.

from __future__ import print_function

from ..config import getConfig, setUsingMCS
from ..tool import GcamTool

PROGRAM = 'gt'

class DummyTool(GcamTool):

    def __init__(self, loadPlugins=False, loadBuiltins=False):
        super(DummyTool, self).__init__(loadPlugins=loadPlugins, loadBuiltins=loadBuiltins)


def _getMainParser():
    '''
    Used only to generate documentation by sphinx argparse extension, in
    which case we don't generate documentation for project-specific plugins.
    '''
    setUsingMCS(True)
    getConfig(reload=True, allowMissing=True)
    tool = DummyTool().getInstance()
    return tool.parser



import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output
from six.moves import shlex_quote

from pygcam.log import getLogger

_logger = getLogger(__name__)


class ActionInfo(object):
    """
    Converts argparse "actions" into HTML widgets.
    """
    ignoredActions = ('_VersionAction', '_HelpAction', '_SubParsersAction')

    def __init__(self, pageId, action, default=None):
        self.action = action
        self.actionClass = type(action).__name__

        opt_strings = action.option_strings
        self.name = ', '.join(opt_strings) if opt_strings else action.dest

        # For constructing command line
        self.option = opt_strings[0] if opt_strings else action.dest
        self.default = default
        self.pageId = pageId
        self.widget = None
        self.value  = None   # set in generated callback
        self.checkDefaultFunc = None # function to decide whether to render option (i.e., not default value)

        t = action.type
        if t == str:
            self.dataType = 'text'
        elif t in (int, float):
            self.dataType = t.__name__
        else:
            self.dataType = 'text'

        self.display = self.actionClass not in self.ignoredActions and self.name != '--VERSION'

        self.layout = None

    def __str__(self):
        return "<ActionInfo %s %s>" % (self.actionClass, self.name)

    def setPageId(self, id):
        self.pageId = id

    def getWidgetId(self):
        return "%s-%s" % (self.pageId, self.action.dest)

    def getWidgetSinkId(self):
        return "_%s-storage_" % self.getWidgetId()

    def generateCallback(self, app):
        """
        Generate a callback function to set this ActionInfo state when the corresponding
        HTML widget is changed. This allows us to interrogate ActionInfo classes for values.

        :param app: a Dash app instance
        :return: none
        """
        inputId  = self.getWidgetId()
        outputId = self.getWidgetSinkId()

        def saveChange(value):
            _logger.debug("Setting value of %s in %s to %r" % (inputId, id(self), value))
            self.value = value
            return value

        _logger.debug("Generating default-setting callback for %s" % inputId)


        # install a callback to the function created above
        app.callback(Output(outputId, 'children'),
                     [Input(inputId, 'value')])(saveChange)

    def radio(self, options, default='', vertical=False):
        self.default = default
        id = self.getWidgetId()

        return dcc.RadioItems(
            id=id,
            options=[{'label': str(opt), 'value': opt} for opt in options],
            value=default if self.value is None else self.value,
            labelStyle={'display': 'block' if vertical else 'inline-block'})

    def input(self, default=None, placeholder='Enter a value...'):
        self.default = self.action.default if default is None else default
        value = self.default if self.value is None else self.value

        if not value and self.action.metavar:
            placeholder = self.action.metavar

        id = self.getWidgetId()
        return dcc.Input(id=id, placeholder=placeholder, type=self.dataType,
                         value=value, className='option')

    def render(self):
        """
        Render an argparse option as HTML
        """
        Y = 'Yes'
        N = 'No'

        actionClass = self.actionClass

        if actionClass == '_StoreTrueAction':
            widget = self.radio([Y, N], default=N)

        elif actionClass == '_StoreFalseAction':
            widget = self.radio([Y, N], default=Y)

        elif actionClass == '_StoreAction':
            choices = self.action.choices
            if choices:
                widget = self.radio(choices, default=self.action.default, vertical=True)
            else:
                widget = self.input(default=self.action.default)

        elif actionClass == '_AppendAction':
            widget = self.input(default=self.action.default)

        elif actionClass == 'ParseCommaList':
            widget = self.input(default=self.action.default)

        else:
            widget = html.P("%s: unknown action type %s" % (self.name, actionClass))

        self.widget = widget

        name = self.name
        if self.action.required:
            name = html.Span(name + '*', className='required')

        help = "(%s) " % self.dataType if self.dataType != 'text' else ''
        help += (self.action.help or '')

        widgetSinkId = self.getWidgetSinkId()

        layout = html.Tr(children=[
            html.Td(name,   className='lined', style={'width': 150}),
            html.Td(widget, className='lined', style={'width': 200}),
            html.Td(help,   className='lined'),

            # invisible sink to store value just so callback is legit
            html.Div(id=widgetSinkId, style={'display': 'none'})
        ], className='cmd')

        self.layout = layout
        return layout

    def cmdlineArg(self):
        """
        Generate the command-line argument corresponding to the user's data entry, or None if
        the default value is set.
        """
        if self.value is None or self.value == self.default:
            return None

        return self._cmdlineArg()

    def _cmdlineArg(self):
        arg = None

        if self.actionClass in ('_StoreTrueAction', '_StoreFalseAction'):
            arg = self.option   # just the flag itself

        elif self.option[0] not in ('-', '+'):
            arg = shlex_quote(self.value)   # positional; just the value

        elif self.value:
            arg = "%s %s" % (self.option, shlex_quote(self.value))

        return arg

class StoreAction(ActionInfo):
    def __init__(self, pageId, action, default):
        super(StoreAction, self).__init__(pageId, action, default)

class StoreConst(ActionInfo):
    def __init__(self, pageId, action, const, default):
        super(StoreConst, self).__init__(pageId, action, default)
        self.const = const

class AppendAction(ActionInfo):
    def __init__(self, pageId, action, default):
        super(AppendAction, self).__init__(pageId, action, default)


def getActionInstance(pageId, action):
    """
    Factory function to create action subclasses
    """
    actionClass = type(action).__name__

    if actionClass in ('_StoreTrueAction', '_StoreFalseAction'):
        obj = StoreConst(pageId, action, const=action.const, default=action.default)

    elif actionClass in ('_StoreAction', 'ParseCommaList'):
        obj = StoreAction(pageId, action, default=action.default)

    elif actionClass == '_AppendAction':
        obj = AppendAction(pageId, action, default=action.default)

    elif actionClass in ActionInfo.ignoredActions:
        obj = None

    else:
        raise Exception('Unrecognized widget action class "%s"' % action)

    return obj


def actionTable(sectionName, actions, helpText=None):
    """
    Generate a table with option names and option data entry widgets
    """
    elements = [html.H3(sectionName, className='band')]
    if helpText:
        elements.append(html.Div(helpText, className='helptext'))

    def th(value):
        return html.Th(value, className='lined')

    headings = html.Tr([th('Option'), th('Value'), th('Description')])
    rows = [headings] + [a.render() for a in actions if a.display]
    elements.append(html.Table(children=rows, className='cmd'))
    return elements



'''
.. Stuff specific to MS Windows
.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from __future__ import print_function
import os
import platform
from errno import EEXIST
from .config import getParamAsBoolean

IsWindows = platform.system() == 'Windows'

if IsWindows:
    # print("Loading Windows support functions...")

    import ctypes
    import win32file
    from pygcam.error import PygcamException

    # From http://stackoverflow.com/questions/8231719/how-to-check-whether-a-file-is-open-and-the-open-status-in-python
    _sopen = ctypes.cdll.msvcrt._sopen
    _close = ctypes.cdll.msvcrt._close
    _SH_DENYRW = 0x10

    def is_open(filename):
        if not os.access(filename, os.F_OK):
            return False  # file doesn't exist
        h = _sopen(filename, 0, _SH_DENYRW, 0)
        if h == 3:
            _close(h)
            return False  # file is not opened by anyone else
        return True  # file is already open

    # Adapted from http://stackoverflow.com/questions/1447575/symlinks-on-windows
    # Win32file is missing this attribute.
    # FILE_ATTRIBUTE_REPARSE_POINT = 0x0400
    #
    # REPARSE_FOLDER = (win32file.FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_REPARSE_POINT)
    #
    # def islinkWindows(path):
    #     """ Windows islink implementation. """
    #     return win32file.GetFileAttributesW(unicode(path)) & REPARSE_FOLDER == REPARSE_FOLDER

    # From http://tomjbward.co.uk/detect-symlink-using-python-2-x-on-windows/
    # def islinkWindows2(path):
    #     return bool(os.path.isdir(path) and \
    #         (win32file.GetFileAttributesW(unicode(path)) & FILE_ATTRIBUTE_REPARSE_POINT))


    # Adapted from http://timgolden.me.uk/python/win32_how_do_i/see_if_two_files_are_the_same_file.html
    def get_read_handle(filename):
        if os.path.isdir(filename):
            dwFlagsAndAttributes = win32file.FILE_FLAG_BACKUP_SEMANTICS
        else:
            dwFlagsAndAttributes = 0

        # CreateFile(fileName, desiredAccess, shareMode, attributes,
        #            CreationDisposition, flagsAndAttributes, hTemplateFile)
        try:
            handle = win32file.CreateFileW(filename,  # with 'W' accepts unicode
                         win32file.GENERIC_READ, win32file.FILE_SHARE_READ, None,
                         win32file.OPEN_EXISTING, dwFlagsAndAttributes, None)
            return handle
        except Exception as e:
            raise PygcamException("get_read_handle(%s) failed: %s" % (filename, e))

    def get_unique_id (hFile):
        (attributes, created_at, accessed_at, written_at,
         volume, file_hi, file_lo, n_links,
         index_hi, index_lo) = win32file.GetFileInformationByHandle(hFile)
        return volume, index_hi, index_lo

    def makeParents(path):
        path = os.path.abspath(path)
        try:
            os.makedirs(os.path.dirname(path), 0o770)
        except OSError as e:
            if e.errno != EEXIST:
                raise

    def samefileWindows(filename1, filename2):
        # If either of the files does not exist, create it and close
        # it to use get_read_handle(), then remove the files we created.
        # Yes, there's a race condition here, and this has the side-effect
        # of possibly creating directories that aren't deleted, but
        # Windows is lame, so we do what we can.
        f1 = f2 = None
        hFile1 = hFile2 = None
        try:

            if not os.path.lexists(filename1):
                makeParents(filename1)
                f1 = open(filename1, 'w')
                f1.close()

            if not os.path.lexists(filename2):
                makeParents(filename2)
                f2 = open(filename2, 'w')
                f2.close()

            hFile1 = get_read_handle(filename1)
            hFile2 = get_read_handle(filename2)
            are_equal = (get_unique_id(hFile1) == get_unique_id(hFile2))

        finally:
            if hFile1:
                hFile1.Close()
            if hFile2:
                hFile2.Close()
            if f1:
                os.unlink(filename1)
            if f2:
                os.unlink(filename2)

        return are_equal

    # Adapted from http://stackoverflow.com/questions/6260149/os-symlink-support-in-windows
    # NOTE: Requires SeCreateSymbolicLink priv.
    def symlinkWindows(src, dst):
        src = src.replace('/', '\\')
        dst = dst.replace('/', '\\')

        if samefileWindows(src, dst):
            raise PygcamException("Attempted to create symlink loop from '%s' to '%s' (same file)" % (dst, src))

        csl = ctypes.windll.kernel32.CreateSymbolicLinkW
        csl.argtypes = (ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_uint32)
        csl.restype = ctypes.c_ubyte

        # links to files and dir differ in Windows
        flags = 1 if os.path.isdir(src) else 0

        try:
            if csl(dst, src, flags) == 0:
                raise ctypes.WinError()
        except Exception as e:
            from pygcam.log import getLogger
            _logger = getLogger(__name__)

            if getParamAsBoolean('GCAM.SymlinkWarning'):
                _logger.error('''
  ============================================================================================================
  WARNING: The current user does not have permission to create symbolic links, forcing pygcam to copy rather
  than symlink files. This uses much more file space than using symlinks, but it works.  To use pygcam more
  efficiently, ask your System Administrator to give you permission to Create Symbol Links (using gpedit.msc)

  See http://superuser.com/questions/104845/permission-to-make-symbolic-links-in-windows-7 for more info.

  Set "GCAM.SymlinkWarning = False" in ~/.pygcam.cfg to suppress this message.
  ============================================================================================================
  ''')
            raise PygcamException("Failed to create symlink '%s' to '%s': %s" % (dst, src, e))

    # The following taken from
    # http://stackoverflow.com/questions/27972776/having-trouble-implementing-a-readlink-function
    #from ctypes import *
    from ctypes.wintypes import DWORD, LPCWSTR, LPVOID, HANDLE, BOOL, USHORT, ULONG, WCHAR
    kernel32 = ctypes.WinDLL('kernel32')
    LPDWORD = ctypes.POINTER(DWORD)
    UCHAR = ctypes.c_ubyte

    GetFileAttributesW = kernel32.GetFileAttributesW
    GetFileAttributesW.restype = DWORD
    GetFileAttributesW.argtypes = (LPCWSTR,) #lpFileName In

    INVALID_FILE_ATTRIBUTES = 0xFFFFFFFF
    FILE_ATTRIBUTE_REPARSE_POINT = 0x00400

    CreateFileW = kernel32.CreateFileW
    CreateFileW.restype = HANDLE
    CreateFileW.argtypes = (LPCWSTR, #lpFileName In
                            DWORD,   #dwDesiredAccess In
                            DWORD,   #dwShareMode In
                            LPVOID,  #lpSecurityAttributes In_opt
                            DWORD,   #dwCreationDisposition In
                            DWORD,   #dwFlagsAndAttributes In
                            HANDLE)  #hTemplateFile In_opt

    CloseHandle = kernel32.CloseHandle
    CloseHandle.restype = BOOL
    CloseHandle.argtypes = (HANDLE,) #hObject In

    INVALID_HANDLE_VALUE = HANDLE(-1).value
    OPEN_EXISTING = 3
    FILE_FLAG_BACKUP_SEMANTICS = 0x02000000
    FILE_FLAG_OPEN_REPARSE_POINT = 0x00200000

    DeviceIoControl = kernel32.DeviceIoControl
    DeviceIoControl.restype = BOOL
    DeviceIoControl.argtypes = (HANDLE,  #hDevice In
                                DWORD,   #dwIoControlCode In
                                LPVOID,  #lpInBuffer In_opt
                                DWORD,   #nInBufferSize In
                                LPVOID,  #lpOutBuffer Out_opt
                                DWORD,   #nOutBufferSize In
                                LPDWORD, #lpBytesReturned Out_opt
                                LPVOID)  #lpOverlapped Inout_opt

    FSCTL_GET_REPARSE_POINT = 0x000900A8
    IO_REPARSE_TAG_MOUNT_POINT = 0xA0000003
    IO_REPARSE_TAG_SYMLINK = 0xA000000C
    MAXIMUM_REPARSE_DATA_BUFFER_SIZE = 0x4000

    class GENERIC_REPARSE_BUFFER(ctypes.Structure):
        _fields_ = (('DataBuffer', UCHAR * 1),)

    class SYMBOLIC_LINK_REPARSE_BUFFER(ctypes.Structure):
        _fields_ = (('SubstituteNameOffset', USHORT),
                    ('SubstituteNameLength', USHORT),
                    ('PrintNameOffset', USHORT),
                    ('PrintNameLength', USHORT),
                    ('Flags', ULONG),
                    ('PathBuffer', WCHAR * 1))
        @property
        def PrintName(self):
            arrayt = WCHAR * (self.PrintNameLength // 2)
            offset = type(self).PathBuffer.offset + self.PrintNameOffset
            return arrayt.from_address(ctypes.addressof(self) + offset).value


    class MOUNT_POINT_REPARSE_BUFFER(ctypes.Structure):
        _fields_ = (('SubstituteNameOffset', USHORT),
                    ('SubstituteNameLength', USHORT),
                    ('PrintNameOffset', USHORT),
                    ('PrintNameLength', USHORT),
                    ('PathBuffer', WCHAR * 1))
        @property
        def PrintName(self):
            arrayt = WCHAR * (self.PrintNameLength // 2)
            offset = type(self).PathBuffer.offset + self.PrintNameOffset
            return arrayt.from_address(ctypes.addressof(self) + offset).value


    class REPARSE_DATA_BUFFER(ctypes.Structure):
        class REPARSE_BUFFER(ctypes.Union):
            _fields_ = (('SymbolicLinkReparseBuffer',
                            SYMBOLIC_LINK_REPARSE_BUFFER),
                        ('MountPointReparseBuffer',
                            MOUNT_POINT_REPARSE_BUFFER),
                        ('GenericReparseBuffer',
                            GENERIC_REPARSE_BUFFER))
        _fields_ = (('ReparseTag', ULONG),
                    ('ReparseDataLength', USHORT),
                    ('Reserved', USHORT),
                    ('ReparseBuffer', REPARSE_BUFFER))
        _anonymous_ = ('ReparseBuffer',)

    REPARSE_FOLDER = (win32file.FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_REPARSE_POINT)

    def islinkWindows(path):
        """ Windows islink implementation. """
        if not os.path.lexists(path):
            return False

        return win32file.GetFileAttributesW(unicode(path)) & REPARSE_FOLDER == REPARSE_FOLDER

    def islinkWindows2(path):
        if not os.path.lexists(path):
            return False

        result = GetFileAttributesW(path)
        if result == INVALID_FILE_ATTRIBUTES:
            raise PygcamException("Can't get file attributes for '%s': %s'" % (path, ctypes.WinError()))

        return bool(result & FILE_ATTRIBUTE_REPARSE_POINT)

    def readlinkWindows(path):
        reparse_point_handle = CreateFileW(path,
                                           0,
                                           0,
                                           None,
                                           OPEN_EXISTING,
                                           FILE_FLAG_OPEN_REPARSE_POINT |
                                           FILE_FLAG_BACKUP_SEMANTICS,
                                           None)
        if reparse_point_handle == INVALID_HANDLE_VALUE:
            from pygcam.log import getLogger
            _logger = getLogger(__name__)
            _logger.info("Can't readlink: %s", path)
            raise ctypes.WinError()

        target_buffer = ctypes.c_buffer(MAXIMUM_REPARSE_DATA_BUFFER_SIZE)
        n_bytes_returned = DWORD()
        io_result = DeviceIoControl(reparse_point_handle,
                                    FSCTL_GET_REPARSE_POINT,
                                    None, 0,
                                    target_buffer, len(target_buffer),
                                    ctypes.byref(n_bytes_returned),
                                    None)
        CloseHandle(reparse_point_handle)
        if not io_result:
            raise ctypes.WinError()

        rdb = REPARSE_DATA_BUFFER.from_buffer(target_buffer)
        if rdb.ReparseTag == IO_REPARSE_TAG_SYMLINK:
            return rdb.SymbolicLinkReparseBuffer.PrintName
        elif rdb.ReparseTag == IO_REPARSE_TAG_MOUNT_POINT:
            return rdb.MountPointReparseBuffer.PrintName

        raise ValueError("not a link")

    # Replace broken functions with those defined above.
    # (In python 2.7.11 os.path.islink() indeed failed to detect link made with mklink)
    os.symlink = symlinkWindows
    os.readlink = readlinkWindows
    os.path.islink = islinkWindows
    os.path.samefile = samefileWindows

def removeSymlink(path):
    """
    On Windows, symlinks to directories  must be removed with os.rmdir(),
    while symlinks to files must be removed with os.remove()
    """
    if not os.path.lexists(path):
        return

    if not os.path.islink(path):
        raise PygcamException("removeSymlink: path '%s' is not a symlink" % path)

    if IsWindows and os.path.isdir(path):
        os.rmdir(path)
    else:
        os.remove(path)




'''
.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from pygcam.constants import GCAM_32_REGIONS

fixedTaxTemplate = '                <fixedTax year="{year}">{tax}</fixedTax>'

regionTemplate = '''        <region name="{region}">
            <ghgpolicy name="CO2">
                <market>{market}</market>
            </ghgpolicy>
        </region>'''

ctaxTemplate = '''<?xml version="1.0" encoding="UTF-8"?>
<scenario name="ctax">
    <world>
        <region name="{region}">
            <ghgpolicy name="CO2">
                <market>{market}</market>
                <isFixedTax>1</isFixedTax>
{taxes}
            </ghgpolicy>
        </region>
{regions}
    </world>
</scenario>
'''

def _futureValue(value, years, rate):
     return value * (1+rate)**years

def _futureValuePairs(value, years, rate):
    start = years[0]
    pairs = [(start, value)]
    for y in years[1:]:
        pairs.append((y, _futureValue(value, y - start, rate)))

    return pairs

def genCarbonTax(value, years, rate, regions=GCAM_32_REGIONS, market='global'):
    '''
    Generate the text of an XML file defining a global carbon tax starting
    at `value` and increasing by `rate` annually. Generate values
    for the give `years`. The first year in `years` is assumed to be
    the year at which the tax starts at `value`.

    :param value: (float) the initial value of the tax ($/tonne)
    :param years: (list(int)) years to set carbon taxes
    :param rate: (float) annual rate of increase.
    :return: (str) the contents of the XML file
    '''
    regionList = []
    firstRegion = regions[0]
    for region in regions[1:]:
        regionList.append(regionTemplate.format(region=region, market=market))

    regionsText = "\n".join(regionList)

    pairs = _futureValuePairs(value, years, rate)
    taxesList = []
    for year, value in pairs:
        taxesList.append(fixedTaxTemplate.format(year=year, tax="%.2f" % value))

    taxes = "\n".join(taxesList)
    fileText = ctaxTemplate.format(taxes=taxes, region=firstRegion, market=market, regions=regionsText)
    return fileText

def genCarbonTaxFile(filename, value, startYear=2020, endYear=2100, timestep=5, rate=0.05,
                     regions=GCAM_32_REGIONS, market='global'):
    '''
    Generate an XML file defining a global carbon tax starting
    at `value` and increasing by `rate` annually. Generate values
    for the years `startYear` to `endYear`, inclusive, and set a new
    tax level every `timestep` years. By default, the tax is applied
    from 2020 to 2100, at 5 year timesteps with an annual increase of 5%.

    :param filename: (str) the path of the XML file to create
    :param value: (float) the initial value of the tax ($/tonne)
    :param startYear: (int) the year in which to establish the carbon tax
    :param endYear: (int) the final year of the carbon tax.
    :param timestep: (int) the number of years after which to update the tax level
    :param rate: (float) annual rate of increase
    :return: none
    '''
    years = range(startYear, endYear + timestep, timestep)
    text = genCarbonTax(value, years, rate, regions=regions, market=market)
    with open(filename, 'w') as f:
        f.write(text)


header = """<?xml version="1.0" encoding="UTF-8"?>
<scenario>
    <world>"""

footer = """
    </world>
</scenario>
"""

ghgPolicyTemplate = """
            <!-- Linked policies must be read in after the policy to which it links
                 This may be difficult to do in some cases so create an empty CO2
                 policy that ensure there is something to link to. The actual policy
                 can be filled in later.
            -->
            <ghgpolicy name="CO2">
                <market>{market}</market>
            </ghgpolicy>"""

firstRegionTemplate = """
                <price-adjust year="1975" fillout="1">{priceAdjust}</price-adjust>
                <demand-adjust year="1975" fillout="1">{demandAdjust}</demand-adjust>
                <price-unit>1990$/tC</price-unit>
                <output-unit>MTC</output-unit>"""

regionTemplate2 = """
        <region name="{region}">{ghgPolicy}
            <linked-ghg-policy name="CO2_LUC">
                <market>{market}</market>
                <linked-policy>CO2</linked-policy>{firstRegionStuff}
            </linked-ghg-policy>
        </region>"""

#
# Function to generate UCT/FFICT linked tax files
#
# (The following note is from Page Kyle of JGCRI)
#
# The price-adjust is defaulted to 0, meaning that no land carbon
# pricing is used (FFICT). To run with land carbon pricing (UCT), it can
# be set to 1, or really any number, and that number will be multiplied
# by the carbon price in order to determine the magnitude of the land
# carbon subsidy. The demand-adjust is also defaulted to 0, and that
# applies in the case of a CO2 emissions cap, or constraint. The
# question here is whether land use change emissions are added to the
# energy and industrial emissions in meeting the constraint. If 0, the
# constraint only applies to the energy and industrial emissions; if set
# to 1, the policy will also consider the LUC emissions. As with the
# price-adjust, it can also be set to any other numerical value.
#
def genLinkedBioCarbonPolicyFile(filename, market='global', regions=None, forTax=True, forCap=False):
    """
    Create the XML for a linked policy to include LUC CO2 in a CO2 cap or tax policy (or both).
    This function generates the equivalent of any of the 4 files in input/policy/:
    global_ffict.xml               (forTax=False, forCap=False)
    global_ffict_in_constraint.xml (forTax=False, forCap=True)
    global_uct.xml                 (forTax=True,  forCap=False)
    global_uct_in_constraint.xml   (forTax=True,  forCap=True)

    However, unlike those files, the market need not be global, and the set of regions to
    which to apply the policy can be specified.

    :param filename: (str) the pathname of the XML file to create
    :param market: (str) the name of the market for which to create the linked policy
    :param regions: (list of str or None) the regions to apply the policy to, or None
      to indicate all regions.
    :param forTax: (bool) True if the linked policy should apply to a CO2 tax
    :param forCap: (bool) True if the linked policy should apply to a CO2 cap
    :return: none
    """
    regions = regions or GCAM_32_REGIONS

    first = True
    parts = [header]

    priceAdjust  = 1.0 if forTax else 0.0
    demandAdjust = 1.0 if forCap else 0.0
    firstRegionStuff = firstRegionTemplate.format(priceAdjust=priceAdjust, demandAdjust=demandAdjust)
    ghgPolicy = ghgPolicyTemplate.format(market=market)

    for region in regions:
        regionElement = regionTemplate2.format(region=region, market=market,
                                               ghgPolicy=ghgPolicy if first else '',
                                               firstRegionStuff=firstRegionStuff if first else '')
        first = False
        parts.append(regionElement)

    parts.append(footer)
    xml = ''.join(parts)

    with open(filename, 'w') as f:
        f.write(xml)


if __name__ == "__main__":
   genCarbonTaxFile('/tmp/ctax.xml',  10, timestep=10, rate=0.07)
   genCarbonTaxFile('/tmp/ctax2.xml', 15, timestep=5,  rate=0.05, regions=['Canada', 'USA', 'EU-15', 'China'], market='region1')






#
# N.B. Avoid loading anything but SubcommandABC at the top level so that
# '-h / --help' runs as quickly as possible. Import needed files in the
# run() method instead.
#
from pygcam.subcommand import SubcommandABC

class MyNewCommand(SubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Short help text for main driver''',
                  'description' : '''Longer description for sub-command'''}

        # The first argument is the name of the new sub-command
        super(MyNewCommand, self).__init__('subCmdName', subparsers, kwargs)

    def addArgs(self):
        '''
        Process the command-line arguments for this sub-command
        '''
        parser = self.parser

        parser.add_argument('-n', '--number', type=int, default=0,
                            help='''A number to demonstrate a command line arg.
                            Replace as needed with your own plugin's args.''')

        return parser

    def run(self, args, tool):
        '''
        Implement the sub-command here. "args" is an `argparse.Namespace` instance
        holding the parsed command-line arguments, and "tool" is a reference to
        the running GcamTool instance.
        '''
        # from pygcam.log import getLogger
        # _logger = getLogger(__name__)

        pass

# An alternative to naming the class 'Plugin' is to assign the class to PluginClass
PluginClass = MyNewCommand



# Created on Mar 20, 2012
#
# @author: Rich Plevin
# @author: Sam Fendell
#
# Copyright (c) 2012-2015. The Regents of the University of California (Regents)
# and Richard Plevin. See the file COPYRIGHT.txt for details.
'''
This module is based on code originally developed by Sam Fendell.
'''
import math
import re
from inspect import getargspec

import numpy as np
from scipy.stats import lognorm, triang, uniform, norm, rv_discrete

from pygcam.log import getLogger
from .error import PygcamMcsUserError

_logger = getLogger(__name__)

def parseDistroKey(key):
    '''
    Gets the name and list of dimensions from a distro key. Inverse of makeDistroKey
    '''
    s = re.split('\[', key)
    return s[0], s[1][:-1].split(',')

def makeDistroKey(name, dimensions, dropZeros=False):
    '''
    Generate a dictionary key for the variable and a list of dimension indices.
    This is a normal function because it is used by both the MatrixRV
    and ParameterSet classes. Inverse of parseDistroKey.
    '''
    if not dimensions:
        return name

    if dropZeros and all(dim == 0 for dim in dimensions):
        return name

    distroKey = name + re.sub(r'[\s\']', r'', str(dimensions))
    return distroKey

# For debugging only
def dumpDistros(distroDict):
    for key in sorted(distroDict.iterkeys()):
        subDistroDict = distroDict[key]
        for d in subDistroDict.values():
            _logger.info(d)

def uniformMinMax(min, max):
    return uniform(loc=min, scale=(max - min))

def uniformRange(range):
    if range <= 0.0:
        raise PygcamMcsUserError("Uniform range must be > 0.0; %f was given" % range)

    return uniformMinMax(-range, range)

def uniformFactor(factor):
    if factor < 0.0 or factor > 1.0:
        raise PygcamMcsUserError("Uniform factor must be between 0.0 and 1.0; %f was given" % factor)

    return uniformMinMax(1 - factor, 1 + factor)

#
# Various ways to specify a lognormal random variable:
#
def lognormalRvForNormal(mu, sigma):
    '''
    Define a lognormal RV by the mean and stdev of the underlying Normal distribution
    '''
    return lognorm(sigma, scale=math.exp(mu))

def lognormalRv(logMean, logStd):
    '''
    Define a lognormal RV by its own mean and stdev
    '''
    logVar = float(logStd) ** 2
    mSqrd = float(logMean) ** 2
    mu = math.log(mSqrd / math.sqrt(logVar + mSqrd))
    sigma = math.sqrt(math.log(logVar / mSqrd + 1))
    return lognormalRvForNormal(mu, sigma)

def lognormalRvFor95th(lo, hi):
    '''
    Define a lognormal RV by its 95% CI.
    '''
    lo = math.log(float(lo))
    hi = math.log(float(hi))
    mu = (lo + hi) / 2.0
    sigma = (hi - mu) / 1.96  # 95th percentile of normal is (+/- 1.96) * sigma
    return lognormalRvForNormal(mu, sigma)

def logfactor(factor):
    if factor < 1.0:
        raise PygcamMcsUserError("LogFactor 'factor' must be >= 1; a value of %f was given." % factor)

    return lognormalRvFor95th(1 / factor, factor)

def triangle(min, mode, max):  # @ReservedAssignment
    scale = max - min
    if scale == 0:
        raise PygcamMcsUserError("Scale of triangle distribution is zero")

    c = (mode - min) / scale  # central value (mode) of the triangle
    return triang(c, loc=min, scale=scale)

def triangleRange(range):
    if range <= 0.0:
        raise PygcamMcsUserError("Triangle range must be between > 0.0; %f was given" % range)

    return triangle(-range, 0, range)

def triangleFactor(factor):
    if factor < 0.0 or factor > 1.0:
        raise PygcamMcsUserError("Triangle factor must be between 0.0 and 1.0; %f was given" % factor)

    return triangle(1 - factor, 1, 1 + factor)

def binary():
    return rv_discrete(name="binary", values=[(0, 1), (0.5, 0.5)])

def integers(min, max):
    min = int(min)
    max = int(max)
    count = max - min + 1
    nums  = range(min, max + 1)
    probs = [1.0/count] * count
    return rv_discrete(name='integers', values=[nums, probs])

class constant():
    """
    Return an object that produces an array holding the given
    constant value. Useful for forcing a parameter to a given value.
    """
    def __init__(self, value):
        self.value = value

    def ppf(self, q):
        n = np.ndarray(len(q))
        n[:] = self.value
        return n

class GridRV(object):
    '''
    Return an object that behaves like an RV in that it returns N values when
    when requested via the ppf (percent point function), though the N values are
    merely a shuffled sequence of a "gridded" range repeated to produce N values.
    No other methods of the standard RV class are implemented. This is intended
    for use in CoreMCS and derivatives only.
    '''
    def __init__(self, min, max, count):
        self.values = np.linspace(min, max, count)
        _logger.debug("Generated values: %s", self.values)


    def ppf(self, q):
        '''
        Return 'n' values from this object's list of values, repeating those values
        as many times as necessary to produce 'n' values, where 'n' is the length of
        the percentile list given by 'q'. (We ignore the values, though.)
        '''
        n = len(q)
        values = self.values
        assert len(values.shape) == 1, "Grid values were converted to ndarray of > 1 dimension"
        count  = values.shape[0]
        reps   = 1 if n <= count else np.ceil(float(n) / count)
        tiled  = np.tile(values, reps)[:n]
        np.random.shuffle(tiled)
        # _logger.debug("tiled=%s", tiled)
        return tiled

class linkedDistro(object):
    def __init__(self, parameter):
        '''Linked to (i.e., shares RV data with) `withParameter`'''
        self.parameter = parameter

    # this is needed to handle linked parameters
    trialData = None

    @classmethod
    def storeTrialData(cls, df):
        cls.trialData = df

    @classmethod
    def getTrialData(cls):
        return cls.trialData

    def ppf(self, q):
        return self.trialData[self.parameter]    # TBD: return as an ndarray

class DistroGen(object):
    '''
    Stores information required to generate a Distro instance from an argDict
    '''
    instances = {}    # Store a dict of our instances internally

    def __init__(self, distName, func):
        self.name = distName
        self.func = func
        self.sig  = DistroGen.signature(distName, getargspec(func).args)
        DistroGen.instances[self.sig] = self

    def __str__(self):
        classname = type(self).__name__
        _logger.debug("<%s dist=%s func=%s sig=%s>", classname, self.distName, self.func, self.sig)

    @classmethod
    def signature(cls, distName, keywords):
        '''
        Makes a unique signature for a distribution type out of its name
        and a collection of argument names.
        '''
        lst = list(keywords)
        lst.append('#' + distName.lower())  # assures that distname doesn't overlap with any of the argument names
        return frozenset(lst)

    @classmethod
    def generator(cls, sig):
        cls.genDistros()
        return cls.instances.get(sig, None)

    def makeRV(self, argDict):
        'Call the generator function with an argDict to create a frozen RV'
        return self.func(**argDict)

    @classmethod
    def genDistros(cls):
        '''
        Generate a basic set of distributions
        '''
        if cls.instances:
            return

        cls('uniform', uniformMinMax)

        # range=0.2 means a Uniform(min=-0.2, max=0.2); used with apply="add"
        cls('uniform', uniformRange)

        # factor=0.2 means Uniform(min=0.8, max=1.2); used with apply="multiply"
        cls('uniform', uniformFactor)

        # LogUniform distribution from 1/n to n, e.g., factor=3 => uniform(1/3, 3)
        cls('loguniform', lambda factor: uniformMinMax(min=1 / factor, max=factor))

        cls('normal', lambda mean, std: norm(loc=mean, scale=std))
        cls('normal', lambda mean, stdev: norm(loc=mean, scale=stdev))          # alternate spelling

        cls('lognormal', lambda mean, std: lognormalRv(mean, std))
        cls('lognormal', lambda low95, high95: lognormalRvFor95th(low95, high95))
        cls('lognormal', logfactor)

        # range=0.2 means a triangle with min, mode, max = (-0.2, 0, +0.2); for apply="add"
        cls('triangle', triangleRange)    # args: range (must be > 0)

        # factor=0.2 means triangle with min, mode, max = (0.8, 1, 1.2); for apply="multiply"
        cls('triangle', triangleFactor)    # args: factor: must be > 0 and < 1

        cls('triangle', triangle)         # args: min, mode, max

        cls('binary', binary)
        cls('integers', integers)     # args: min, max (inclusive)

        # Gridded (non-random) sequence.
        # Returns a frozen RV-like object with a "ppf" method that returns a sequence of values
        # produced by cycling through 'count' values evenly spaced starting at 'min' and ending
        # at 'max'.
        cls('grid', lambda min, max, count: GridRV(min, max, count))

        # TBD: Test this
        cls('constant', lambda value: constant(value))

        cls('linked', lambda parameter: linkedDistro(parameter))       # TBD: could be generalized



#!/usr/bin/env python2

from pygcam import tool
from pygcam.config import setUsingMCS

# Turn this off for pyinstaller-run versions since MCS works only
# Linux and pyinstaller is currently used only for Mac and Windows.
setUsingMCS(False)

tool.main()



import os
import pandas as pd
from pygcam.log import getLogger
from pygcam.xmlEditor import XMLEditor, callableMethod

_logger = getLogger(__name__)

class MyCustomClass(XMLEditor):
    '''
    XML editing subclass to extend XML setup functionality
    '''
    @callableMethod
    def updateEnergyCosts(self, startYear, endYear, techs=None):
        """
        Use updated non-energy-costs provided by Matteo.
        """
        techs = techs or ('oil refining', 'coal to liquids', 'gas to liquids',
                          'corn ethanol', 'biodiesel', 'sugar cane ethanol',
                          'cellulosic ethanol', 'FT biofuels')

        _logger.debug('Updating energy costs for %s, %s-%s', techs, startYear, endYear)

        def getCost(df, tech):
            query = 'technology=="%s" and year >= %d and year <= %d' % (tech, startYear, endYear)
            df = df.query(query)
            df.set_index('year', inplace=True)
            costs = df['input.cost']
            return costs

        filename = os.path.join(self.xmlSourceDir, 'baseline', 'L222.GlobalTechCost_en-Modified.csv')
        df = pd.read_csv(filename, skiprows=4)

        for tech in techs:
            # The three fossil technologies are in their own subsectors of the
            # same name. The rest are biomass liquids.
            subSector = tech if tech in ['oil refining', 'coal to liquids', 'gas to liquids'] else 'biomass liquids'
            costs = getCost(df, tech)
            self.setGlobalTechNonEnergyCost('refining', subSector, tech, costs)

